"""
XML utility functions and classes for XBlocks.
Note: Most of the functionality is taken from the edx-platform's XmlMixin.
https://github.com/openedx/edx-platform/blob/master/xmodule/xml_block.py
"""

import copy
import datetime
import json
import logging
import os
from typing import Any

from django.core.serializers.json import DjangoJSONEncoder
from lxml import etree
from lxml.etree import ElementTree, XMLParser, _Element
from opaque_keys.edx.keys import CourseKey, UsageKey
from xblock.core import XML_NAMESPACES, XBlock
from xblock.fields import Dict, Field, Scope, ScopeIds

log = logging.getLogger(__name__)

# assume all XML files are persisted as utf-8.
EDX_XML_PARSER = XMLParser(dtd_validation=False, load_dtd=False, remove_blank_text=True, encoding='utf-8')


class EdxJSONEncoder(DjangoJSONEncoder):
    """
    Custom JSONEncoder that handles ``Location`` and ``datetime.datetime`` objects.
    Encodes ``Location`` as its URL string form, and ``datetime.datetime`` as an ISO 8601 string.
    """

    def default(self, o):
        if isinstance(o, (CourseKey, UsageKey)):
            return str(o)
        elif isinstance(o, datetime.datetime):
            if o.tzinfo is not None:
                if o.utcoffset() is None:
                    return o.isoformat() + "Z"
                else:
                    return o.isoformat()
            else:
                return o.isoformat()
        else:
            return super().default(o)


def name_to_pathname(name: str) -> str:
    """
    Convert a location name for use in a path: replace ':' with '/'.
    This allows users of the xml format to organize content into directories
    """
    return name.replace(':', '/')


def is_pointer_tag(xml_obj: _Element) -> bool:
    """
    Check if xml_obj is a pointer tag: <blah url_name="something" />.
    No children, one attribute named url_name, no text.

    Special case for course roots: the pointer is
      <course url_name="something" org="myorg" course="course">

    xml_obj: an etree Element

    Returns a bool.
    """
    if xml_obj.tag != "course":
        expected_attr = {'url_name'}
    else:
        expected_attr = {'url_name', 'course', 'org'}

    actual_attr = set(xml_obj.attrib.keys())

    has_text = xml_obj.text is not None and len(xml_obj.text.strip()) > 0

    return len(xml_obj) == 0 and actual_attr == expected_attr and not has_text


def serialize_field(value: Any) -> str:
    """
    Return a string version of the value (where value is the JSON-formatted, internally stored value).

    If the value is a string, then we simply return what was passed in.
    Otherwise, we return json.dumps on the input value.
    """
    if isinstance(value, str):
        return value
    elif isinstance(value, datetime.datetime):
        if value.tzinfo is not None and value.utcoffset() is None:
            return value.isoformat() + 'Z'
        return value.isoformat()

    return json.dumps(value, cls=EdxJSONEncoder)


def deserialize_field(field: Field, value: str) -> Any:
    """
    Deserialize the string version to the value stored internally.

    Note that this is not the same as the value returned by from_json, as model types typically store
    their value internally as JSON. By default, this method will return the result of calling json.loads
    on the supplied value, unless json.loads throws a TypeError, or the type of the value returned by json.loads
    is not supported for this class (from_json throws an Error). In either of those cases, this method returns
    the input value.
    """
    try:
        deserialized = json.loads(value)
        if deserialized is None:
            return deserialized
        try:
            field.from_json(deserialized)
            return deserialized
        except (ValueError, TypeError):
            # Support older serialized version, which was just a string, not result of json.dumps.
            # If the deserialized version cannot be converted to the type (via from_json),
            # just return the original value. For example, if a string value of '3.4' was
            # stored for a String field (before we started storing the result of json.dumps),
            # then it would be deserialized as 3.4, but 3.4 is not supported for a String
            # field. Therefore field.from_json(3.4) will throw an Error, and we should
            # actually return the original value of '3.4'.
            return value

    except (ValueError, TypeError):
        # Support older serialized version.
        return value


def own_metadata(block: XBlock) -> dict[str, Any]:
    """
    Return a JSON-friendly dictionary that contains only non-inherited field
    keys, mapped to their serialized values
    """
    result = {}
    for field in block.fields.values():  # lint-amnesty, pylint: disable=no-member
        if field.scope == Scope.settings and field.is_set_on(block):
            try:
                result[field.name] = field.read_json(block)
            except TypeError as exception:
                exception_message = "{message}, Block-location:{location}, Field-name:{field_name}".format(
                    message=str(exception),
                    location=str(block.location),
                    field_name=field.name
                )
                raise TypeError(exception_message)  # lint-amnesty, pylint: disable=raise-missing-from
    return result


class LegacyXmlMixin:
    """
    Class containing XML parsing functionality for the extracted XBlocks.
    NOTE: This is a temporary solution, we don't want new XBlocks to use this mixin.
    It will be removed in the future improvements.
    """
    # Extension to append to filename paths
    filename_extension = 'xml'

    xml_attributes = Dict(help="Map of unhandled xml attributes, used only for storage between import and export",
                          default={}, scope=Scope.settings)

    metadata_to_strip = ('data_dir',
                         'tabs', 'grading_policy',
                         'discussion_blackouts',
                         # VS[compat]
                         # These attributes should have been removed from here once all 2012-fall courses imported into
                         # the CMS and "inline" OLX format deprecated. But, it never got deprecated. Moreover, it's
                         # widely used to this date. So, we still have to strip them. Also, removing of "filename"
                         # changes OLX returned by `/api/olx-export/v1/xblock/{block_id}/`, which indicates that some
                         # places in the platform rely on it.
                         'course', 'org', 'url_name', 'filename',
                         # Used for storing xml attributes between import and export, for roundtrips
                         'xml_attributes',
                         # Used by _import_xml_node_to_parent in cms/djangoapps/contentstore/helpers.py to prevent
                         # XmlMixin from treating some XML nodes as "pointer nodes".
                         "x-is-pointer-node",
                         )

    # This is a categories to fields map that contains the block category specific fields which should not be
    # cleaned and/or override while adding xml to node.
    metadata_to_not_to_clean = {
        # A category `video` having `sub` and `transcripts` fields
        # which should not be cleaned/override in an xml object.
        'video': ('sub', 'transcripts')
    }

    metadata_to_export_to_policy = ('discussion_topics',)

    @staticmethod
    def _get_metadata_from_xml(xml_object, remove=True):
        """
        Extract the metadata from the XML.
        """
        meta = xml_object.find('meta')
        if meta is None:
            return ''
        dmdata = meta.text
        if remove:
            xml_object.remove(meta)
        return dmdata

    @classmethod
    def definition_from_xml(cls, xml_object, system):
        """
        Return the definition to be passed to the newly created block
        during from_xml

        xml_object: An etree Element
        """
        raise NotImplementedError("%s does not implement definition_from_xml" % cls.__name__)

    @classmethod
    def clean_metadata_from_xml(cls, xml_object, excluded_fields=()):
        """
        Remove any attribute named for a field with scope Scope.settings from the supplied
        xml_object
        """
        for field_name, field in cls.fields.items():
            if (field.scope == Scope.settings
                    and field_name not in excluded_fields
                    and xml_object.get(field_name) is not None):
                del xml_object.attrib[field_name]

    @classmethod
    def file_to_xml(cls, file_object):
        """
        Used when this module wants to parse a file object to xml
        that will be converted to the definition.

        Returns an lxml Element
        """
        return etree.parse(file_object, parser=EDX_XML_PARSER).getroot()

    @classmethod
    def load_file(cls, filepath, fs, def_id):  # pylint: disable=invalid-name
        """
        Open the specified file in fs, and call cls.file_to_xml on it,
        returning the lxml object.

        Add details and reraise on error.
        """
        try:
            with fs.open(filepath) as xml_file:
                return cls.file_to_xml(xml_file)
        except Exception as err:
            # Add info about where we are, but keep the traceback
            raise Exception(f'Unable to load file contents at path {filepath} for item {def_id}: {err}') from err

    @classmethod
    def load_definition(cls, xml_object, system, def_id, id_generator):
        """
        Load a block from the specified xml_object.
        Subclasses should not need to override this except in special
        cases (e.g. html block)

        Args:
            xml_object: an lxml.etree._Element containing the definition to load
            system: the modulestore system (aka, runtime) which accesses data and provides access to services
            def_id: the definition id for the block--used to compute the usage id and asides ids
            id_generator: used to generate the usage_id
        """

        # VS[compat]
        # The filename attr should have been removed once all 2012-fall courses imported into the CMS and "inline" OLX
        # format deprecated. This never happened, and `filename` is still used, so we have too keep both formats.
        filename = xml_object.get('filename')
        if filename is None:
            definition_xml = copy.deepcopy(xml_object)
            filepath = ''
            aside_children = []
        else:
            filepath = cls._format_filepath(xml_object.tag, filename)

            # VS[compat]
            # If the file doesn't exist at the right path, give the class a chance to fix it up. The file will be
            # written out again in the correct format. This should have gone away once the CMS became online and had
            # imported all 2012-fall courses from XML.
            if not system.resources_fs.exists(filepath) and hasattr(cls, 'backcompat_paths'):
                candidates = cls.backcompat_paths(filepath)
                for candidate in candidates:
                    if system.resources_fs.exists(candidate):
                        filepath = candidate
                        break

            definition_xml = cls.load_file(filepath, system.resources_fs, def_id)
            usage_id = id_generator.create_usage(def_id)
            aside_children = system.parse_asides(definition_xml, def_id, usage_id, id_generator)

            # Add the attributes from the pointer node
            definition_xml.attrib.update(xml_object.attrib)

        definition_metadata = cls._get_metadata_from_xml(definition_xml)
        cls.clean_metadata_from_xml(definition_xml)
        definition, children = cls.definition_from_xml(definition_xml, system)
        if definition_metadata:
            definition['definition_metadata'] = definition_metadata
        definition['filename'] = [filepath, filename]

        if aside_children:
            definition['aside_children'] = aside_children

        return definition, children

    @classmethod
    def load_metadata(cls, xml_object):
        """
        Read the metadata attributes from this xml_object.

        Returns a dictionary {key: value}.
        """
        metadata = {'xml_attributes': {}}
        for attr, val in xml_object.attrib.items():

            if attr in cls.metadata_to_strip:
                # don't load these
                continue

            if attr not in cls.fields:
                metadata['xml_attributes'][attr] = val
            else:
                metadata[attr] = deserialize_field(cls.fields[attr], val)
        return metadata

    @classmethod
    def apply_policy(cls, metadata, policy):
        """
        Add the keys in policy to metadata, after processing them
        through the attrmap.  Updates the metadata dict in place.
        """
        for attr, value in policy.items():
            if attr not in cls.fields:
                # Store unknown attributes coming from policy.json
                # in such a way that they will export to xml unchanged
                metadata['xml_attributes'][attr] = value
            else:
                metadata[attr] = value

    @classmethod
    def parse_xml(cls, node, runtime, keys):
        """
        Use `node` to construct a new block.

        Arguments:
            node (etree.Element): The xml node to parse into an xblock.

            runtime (:class:`.Runtime`): The runtime to use while parsing.

            keys (:class:`.ScopeIds`): The keys identifying where this block
                will store its data.

        Returns (XBlock): The newly parsed XBlock

        """
        if keys is None:
            # Passing keys=None is against the XBlock API but some platform tests do it.
            def_id = runtime.id_generator.create_definition(node.tag, node.get('url_name'))
            keys = ScopeIds(None, node.tag, def_id, runtime.id_generator.create_usage(def_id))
        aside_children = []

        # Let the runtime construct the block. It will have a proper, inheritance-aware field data store.
        xblock = runtime.construct_xblock_from_class(cls, keys)

        # VS[compat]
        # In 2012, when the platform didn't have CMS, and all courses were handwritten XML files, problem tags
        # contained XML problem descriptions withing themselves. Later, when Studio has been created, and "pointer" tags
        # became the preferred problem format, edX has to add this compatibility code to 1) support both pre- and
        # post-Studio course formats simulteneously, and 2) be able to migrate 2012-fall courses to Studio. Old style
        # support supposed to be removed, but the deprecation process have never been initiated, so this
        # compatibility must stay, probably forever.
        if is_pointer_tag(node):
            # new style:
            # read the actual definition file--named using url_name.replace(':','/')
            definition_xml, filepath = cls.load_definition_xml(node, runtime, keys.def_id)
            aside_children = runtime.parse_asides(definition_xml, keys.def_id, keys.usage_id, runtime.id_generator)
        else:
            filepath = None
            definition_xml = node

        # Note: removes metadata.
        definition, children = cls.load_definition(definition_xml, runtime, keys.def_id, runtime.id_generator)

        # VS[compat]
        # Make Ike's github preview links work in both old and new file layouts.
        if is_pointer_tag(node):
            # new style -- contents actually at filepath
            definition['filename'] = [filepath, filepath]

        metadata = cls.load_metadata(definition_xml)

        # move definition metadata into dict
        dmdata = definition.get('definition_metadata', '')
        if dmdata:
            metadata['definition_metadata_raw'] = dmdata
            try:
                metadata.update(json.loads(dmdata))
            except Exception as err:  # lint-amnesty, pylint: disable=broad-except
                log.debug('Error in loading metadata %r', dmdata, exc_info=True)
                metadata['definition_metadata_err'] = str(err)

        definition_aside_children = definition.pop('aside_children', None)
        if definition_aside_children:
            aside_children.extend(definition_aside_children)

        # Set/override any metadata specified by policy
        cls.apply_policy(metadata, runtime.get_policy(keys.usage_id))

        # We don't want to use InheritanceKeyValueStore here (doing so will necessitate a lot of code to be moved
        # from the `xmodule.modulestore.inheritance` into this class), so we manually set the fields on the block.
        field_data = {**metadata, **definition}

        for field_name, value in field_data.items():
            # The 'xml_attributes' field has a special setter logic in its Field class,
            # so we must handle it carefully to avoid duplicating data.
            if field_name == "xml_attributes":
                # The 'filename' attribute is specially handled for git links.
                value["filename"] = definition.get("filename", ["", None])
                xblock.xml_attributes.update(value)
            elif field_name in xblock.fields:
                setattr(xblock, field_name, value)

        xblock.children = children

        if aside_children:
            cls.add_applicable_asides_to_block(xblock, runtime, aside_children)

        return xblock

    @classmethod
    def add_applicable_asides_to_block(cls, block, runtime, aside_children):
        """
        Add asides to the block. Moved this out of the parse_xml method to use it in the VideoBlock.parse_xml
        """
        asides_tags = [aside_child.tag for aside_child in aside_children]
        asides = runtime.get_asides(block)
        for aside in asides:
            if aside.scope_ids.block_type in asides_tags:
                block.add_aside(aside)

    @classmethod
    def parse_xml_new_runtime(cls, node, runtime, keys):
        """
        This XML lives within Learning Core and the new runtime doesn't need this
        legacy XModule code. Use the "normal" XBlock parsing code.
        """
        try:
            return super().parse_xml_new_runtime(node, runtime, keys)
        except AttributeError:
            return super().parse_xml(node, runtime, keys)

    @classmethod
    def load_definition_xml(cls, node, runtime, def_id):
        """
        Loads definition_xml stored in a dedicated file
        """
        url_name = node.get('url_name')
        filepath = cls._format_filepath(node.tag, name_to_pathname(url_name))
        definition_xml = cls.load_file(filepath, runtime.resources_fs, def_id)
        return definition_xml, filepath

    @classmethod
    def _format_filepath(cls, category, name):
        return f'{category}/{name}.{cls.filename_extension}'

    def export_to_file(self):
        """If this returns True, write the definition of this block to a separate
        file.

        NOTE: Do not override this without a good reason.  It is here
        specifically for customtag...
        """
        return True

    def add_xml_to_node(self, node):
        """
        For exporting, set data on `node` from ourselves.
        """
        xml_object = self.definition_to_xml(self.runtime.export_fs)

        # If xml_object is None, we don't know how to serialize this node, but
        # we shouldn't crash out the whole export for it.
        if xml_object is None:
            return

        for aside in self.runtime.get_asides(self):
            if aside.needs_serialization():
                aside_node = etree.Element("unknown_root", nsmap=XML_NAMESPACES)
                aside.add_xml_to_node(aside_node)
                xml_object.append(aside_node)

        not_to_clean_fields = self.metadata_to_not_to_clean.get(self.category, ())
        self.clean_metadata_from_xml(xml_object, excluded_fields=not_to_clean_fields)

        # Set the tag on both nodes so we get the file path right.
        xml_object.tag = self.category
        node.tag = self.category

        # Add the non-inherited metadata
        for attr in sorted(own_metadata(self)):
            # don't want e.g. data_dir
            if (attr not in self.metadata_to_strip
                    and attr not in self.metadata_to_export_to_policy
                    and attr not in not_to_clean_fields):
                val = serialize_field(self.fields[attr].to_json(getattr(self, attr)))
                try:
                    xml_object.set(attr, val)
                except Exception:  # lint-amnesty, pylint: disable=broad-except
                    logging.exception(
                        'Failed to serialize metadata attribute %s with value %s in module %s. '
                        'This could mean data loss!!!',
                        attr, val, self.url_name
                    )

        for key, value in self.xml_attributes.items():
            if key not in self.metadata_to_strip:
                xml_object.set(key, serialize_field(value))

        if self.export_to_file():
            # Write the definition to a file
            url_path = name_to_pathname(self.url_name)
            filepath = self._format_filepath(self.category, url_path)
            self.runtime.export_fs.makedirs(os.path.dirname(filepath), recreate=True)
            with self.runtime.export_fs.open(filepath, 'wb') as fileobj:
                ElementTree(xml_object).write(fileobj, pretty_print=True, encoding='utf-8')
        else:
            # Write all attributes from xml_object onto node
            node.clear()
            node.tag = xml_object.tag
            node.text = xml_object.text
            node.tail = xml_object.tail
            node.attrib.update(xml_object.attrib)
            node.extend(xml_object)

        # Do not override an existing value for the course.
        if not node.get('url_name'):
            node.set('url_name', self.url_name)

        # We do not need to cater the `course` category here in xblocks_contrib,
        # because course export is handled in the edx-platform code.

        # Special case for course pointers:
        # if self.category == 'course':
        #     # add org and course attributes on the pointer tag
        #     node.set('org', self.location.org)
        #     node.set('course', self.location.course)

    def definition_to_xml(self, resource_fs):
        """
        Return a new etree Element object created from this modules definition.
        """
        raise NotImplementedError(
            "%s does not implement definition_to_xml" % self.__class__.__name__)

    @property
    def non_editable_metadata_fields(self):
        """
        Return a list of all metadata fields that cannot be edited.
        """
        non_editable_fields = super().non_editable_metadata_fields
        non_editable_fields.append(LegacyXmlMixin.xml_attributes)
        return non_editable_fields



"""Poll block is ungraded xmodule used by students to
to do set of polls.

On the client side we show:
If student does not yet anwered - Question with set of choices.
If student have answered - Question with statistics for each answers.
"""

import copy
import datetime
import json
import logging
from collections import OrderedDict
from copy import deepcopy

import markupsafe
from django.utils.translation import gettext_noop as _
from lxml import etree
from opaque_keys.edx.keys import UsageKey
from web_fragments.fragment import Fragment
from xblock.core import XML_NAMESPACES, XBlock
from xblock.fields import Boolean, Dict, List, Scope, ScopeIds, String
from xblock.utils.resources import ResourceLoader

Text = markupsafe.escape
resource_loader = ResourceLoader(__name__)
log = logging.getLogger(__name__)

# assume all XML files are persisted as utf-8.
EDX_XML_PARSER = etree.XMLParser(dtd_validation=False, load_dtd=False, remove_blank_text=True, encoding="utf-8")


def name_to_pathname(name):
    """
    Convert a location name for use in a path: replace ':' with '/'.
    This allows users of the xml format to organize content into directories
    """
    return name.replace(":", "/")


def is_pointer_tag(xml_obj):
    """
    Check if xml_obj is a pointer tag: <blah url_name="something" />.
    No children, one attribute named url_name, no text.

    Special case for course roots: the pointer is
      <course url_name="something" org="myorg" course="course">

    xml_obj: an etree Element

    Returns a bool.
    """
    if xml_obj.tag != "course":
        expected_attr = {"url_name"}
    else:
        expected_attr = {"url_name", "course", "org"}

    actual_attr = set(xml_obj.attrib.keys())

    has_text = xml_obj.text is not None and len(xml_obj.text.strip()) > 0

    return len(xml_obj) == 0 and actual_attr == expected_attr and not has_text


def serialize_field(value):
    """
    Return a string version of the value (where value is the JSON-formatted, internally stored value).

    If the value is a string, then we simply return what was passed in.
    Otherwise, we return json.dumps on the input value.
    """
    if isinstance(value, str):
        return value
    elif isinstance(value, datetime.datetime):
        if value.tzinfo is not None and value.utcoffset() is None:
            return value.isoformat() + "Z"
        return value.isoformat()

    return json.dumps(value)


def deserialize_field(field, value):
    """
    Deserialize the string version to the value stored internally.

    Note that this is not the same as the value returned by from_json, as model types typically store
    their value internally as JSON. By default, this method will return the result of calling json.loads
    on the supplied value, unless json.loads throws a TypeError, or the type of the value returned by json.loads
    is not supported for this class (from_json throws an Error). In either of those cases, this method returns
    the input value.
    """
    try:
        deserialized = json.loads(value)
        if deserialized is None:
            return deserialized
        try:
            field.from_json(deserialized)
            return deserialized
        except (ValueError, TypeError):
            # Support older serialized version, which was just a string, not result of json.dumps.
            # If the deserialized version cannot be converted to the type (via from_json),
            # just return the original value. For example, if a string value of '3.4' was
            # stored for a String field (before we started storing the result of json.dumps),
            # then it would be deserialized as 3.4, but 3.4 is not supported for a String
            # field. Therefore field.from_json(3.4) will throw an Error, and we should
            # actually return the original value of '3.4'.
            return value

    except (ValueError, TypeError):
        # Support older serialized version.
        return value


def HTML(html):  # pylint: disable=invalid-name
    """
    Mark a string as already HTML, so that it won't be escaped before output.

    Use this function when formatting HTML into other strings.  It must be
    used in conjunction with ``Text()``, and both ``HTML()`` and ``Text()``
    must be closed before any calls to ``format()``::

        <%page expression_filter="h"/>
        <%!
        from django.utils.translation import gettext as _

        from openedx.core.djangolib.markup import HTML, Text
        %>
        ${Text(_("Write & send {start}email{end}")).format(
            start=HTML("<a href='mailto:{}'>").format(user.email),
            end=HTML("</a>"),
        )}

    """
    return markupsafe.Markup(html)


def stringify_children(node):
    """
    Return all contents of an xml tree, without the outside tags.
    e.g. if node is parse of

        "<html a="b" foo="bar">Hi <div>there <span>Bruce</span><b>!</b></div><html>"

    should return

        "Hi <div>there <span>Bruce</span><b>!</b></div>"

    fixed from
    http://stackoverflow.com/questions/4624062/get-all-text-inside-a-tag-in-lxml
    """
    # Useful things to know:

    # node.tostring() -- generates xml for the node, including start
    #                 and end tags.  We'll use this for the children.
    # node.text -- the text after the end of a start tag to the start
    #                 of the first child
    # node.tail -- the text after the end this tag to the start of the
    #                 next element.
    parts = [node.text]
    for c in node.getchildren():
        parts.append(etree.tostring(c, with_tail=True, encoding="unicode"))

    # filter removes possible Nones in texts and tails
    return "".join([part for part in parts if part])


@XBlock.needs("i18n")
class PollBlock(XBlock):
    """Poll Block"""

    # Indicates that this XBlock has been extracted from edx-platform.
    is_extracted = True

    # Name of poll to use in links to this poll
    display_name = String(
        help=_("The display name for this component."),
        scope=Scope.settings,
        default="poll_question",
    )

    voted = Boolean(help=_("Whether this student has voted on the poll"), scope=Scope.user_state, default=False)
    poll_answer = String(help=_("Student answer"), scope=Scope.user_state, default="")
    poll_answers = Dict(help=_("Poll answers from all students"), scope=Scope.user_state_summary)

    # List of answers, in the form {'id': 'some id', 'text': 'the answer text'}
    answers = List(help=_("Poll answers from xml"), scope=Scope.content, default=[])

    question = String(help=_("Poll question"), scope=Scope.content, default="")

    # This is a categories to fields map that contains the block category specific fields which should not be
    # cleaned and/or override while adding xml to node.
    metadata_to_not_to_clean = {
        # A category `video` having `sub` and `transcripts` fields
        # which should not be cleaned/override in an xml object.
        "video": ("sub", "transcripts")
    }
    metadata_to_export_to_policy = ("discussion_topics",)
    js_module_name = "poll"

    # Extension to append to filename paths
    filename_extension = "xml"

    xml_attributes = Dict(
        help="Map of unhandled xml attributes, used only for storage between import and export",
        default={},
        scope=Scope.settings,
    )

    metadata_to_strip = (
        "data_dir",
        "tabs",
        "grading_policy",
        "discussion_blackouts",
        # VS[compat]
        # These attributes should have been removed from here once all 2012-fall courses imported into
        # the CMS and "inline" OLX format deprecated. But, it never got deprecated. Moreover, it's
        # widely used to this date. So, we still have to strip them. Also, removing of "filename"
        # changes OLX returned by `/api/olx-export/v1/xblock/{block_id}/`, which indicates that some
        # places in the platform rely on it.
        "course",
        "org",
        "url_name",
        "filename",
        # Used for storing xml attributes between import and export, for roundtrips
        "xml_attributes",
        # Used by _import_xml_node_to_parent in cms/djangoapps/contentstore/helpers.py to prevent
        # XmlMixin from treating some XML nodes as "pointer nodes".
        "x-is-pointer-node",
    )

    _tag_name = "poll_question"
    _child_tag_name = "answer"

    @property
    def xblock_kvs(self):
        """
        Retrieves the internal KeyValueStore for this XModule.

        Should only be used by the persistence layer. Use with caution.
        """
        # if caller wants kvs, caller's assuming it's up to date; so, decache it
        self.save()
        return self._field_data._kvs  # pylint: disable=protected-access

    @property
    def url_name(self):
        return self.location.block_id

    @property
    def course_id(self):
        return self.location.course_key

    @property
    def category(self):
        return self.scope_ids.block_type

    @property
    def location(self):
        return self.scope_ids.usage_id

    @location.setter
    def location(self, value):
        assert isinstance(value, UsageKey)
        self.scope_ids = self.scope_ids._replace(
            def_id=value,  # Note: assigning a UsageKey as def_id is OK in old mongo / import system but wrong in split
            usage_id=value,
        )

    def handle_ajax(self, dispatch, data):  # legacy support for tests
        """
        Legacy method to mimic old ajax handler behavior for backward compatibility.
        """
        if dispatch == "get_state":
            return json.dumps(self.handle_get_state(data))
        else:
            return json.dumps(self.submit_answer(dispatch))

    def student_view(self, _context):
        """
        Renders the student view.
        """

        frag = Fragment()
        frag.add_content(
            resource_loader.render_django_template(
                "templates/poll.html",
                {
                    "element_id": self.scope_ids.usage_id.html_id(),
                    "element_class": self.scope_ids.usage_id.block_type,
                    "configuration_json": self.dump_poll(),
                },
                i18n_service=self.runtime.service(self, "i18n"),
            )
        )
        frag.add_css(resource_loader.load_unicode("static/css/poll.css"))
        frag.add_javascript(resource_loader.load_unicode("static/js/src/poll.js"))
        frag.initialize_js("PollBlock")
        return frag

    def dump_poll(self):
        """Dump poll information.

        Returns:
            string - Serialize json.
        """
        # FIXME: hack for resolving caching `default={}` during definition
        # poll_answers field

        if self.poll_answers is None:
            self.poll_answers = {}

        answers_to_json = OrderedDict()

        # # # FIXME: fix this, when xblock support mutable types.
        # # # Now we use this hack.
        temp_poll_answers = self.poll_answers

        # # # Fill self.poll_answers, prepare data for template context.
        for answer in self.answers:
            # Set default count for answer = 0.
            if answer["id"] not in temp_poll_answers:
                temp_poll_answers[answer["id"]] = 0
            answers_to_json[answer["id"]] = answer["text"]
        self.poll_answers = temp_poll_answers

        return json.dumps(
            {
                "answers": answers_to_json,
                "question": self.question,
                "poll_answer": self.poll_answer,
                "poll_answers": self.poll_answers,
                "total": sum(self.poll_answers.values()) if self.voted else 0,
                "reset": str(self.xml_attributes.get("reset", "true")).lower(),
            }
        )

    @XBlock.json_handler
    def handle_get_state(self, data, suffix=""):  # pylint: disable=unused-argument
        return {
            "poll_answer": self.poll_answer,
            "poll_answers": self.poll_answers,
            "total": sum(self.poll_answers.values()),
        }

    def submit_answer(self, answer):
        """
        Submits a poll answer.
        """
        if not answer:
            return {"error": "No answer provided!"}

        if answer in self.poll_answers and not self.voted:
            # FIXME: fix this, when xblock will support mutable types.
            # Now we use this hack.
            temp_poll_answers = self.poll_answers
            temp_poll_answers[answer] += 1
            self.poll_answers = temp_poll_answers

            self.voted = True
            self.poll_answer = answer
            return {
                "poll_answers": self.poll_answers,
                "total": sum(self.poll_answers.values()),
                "callback": {"objectName": "Conditional"},
            }
        return {"error": "Unknown Command!"}

    @XBlock.json_handler
    def handle_submit_state(self, data, suffix=""):  # pylint: disable=unused-argument
        """
        handler to submit poll answer.
        """
        answer = data.get("answer")  # Extract the answer from the data payload
        return self.submit_answer(answer)

    @XBlock.json_handler
    def handle_reset_state(self):
        """
        handler to Reset poll answer.
        """

        self.voted = False

        # FIXME: fix this, when xblock will support mutable types.
        # Now we use this hack.
        temp_poll_answers = self.poll_answers
        temp_poll_answers[self.poll_answer] -= 1
        self.poll_answers = temp_poll_answers
        self.poll_answer = ""
        return {"status": "success"}

    @staticmethod
    def workbench_scenarios():
        """Create canned scenario for display in the workbench."""
        return [
            (
                "PollBlock",
                """<_poll_question_extracted/>
                """,
            ),
            (
                "Multiple PollBlock",
                """<vertical_demo>
                <_poll_question_extracted/>
                <_poll_question_extracted/>
                <_poll_question_extracted/>
                </vertical_demo>
                """,
            ),
        ]

    def get_explicitly_set_fields_by_scope(self, scope=Scope.content):
        """
        Get a dictionary of the fields for the given scope which are set explicitly on this xblock. (Including
        any set to None.)
        """
        result = {}
        for field in self.fields.values():
            if field.scope == scope and field.is_set_on(self):
                try:
                    result[field.name] = field.read_json(self)
                except TypeError as exception:
                    exception_message = "{message}, Block-location:{location}, Field-name:{field_name}".format(
                        message=str(exception), location=str(self.location), field_name=field.name
                    )
                    raise TypeError(exception_message)  # pylint: disable=raise-missing-from
        return result

    @staticmethod
    def _get_metadata_from_xml(xml_object, remove=True):
        """
        Extract the metadata from the XML.
        """
        meta = xml_object.find("meta")
        if meta is None:
            return ""
        dmdata = meta.text
        if remove:
            xml_object.remove(meta)
        return dmdata

    @classmethod
    def definition_from_xml(cls, xml_object, system):  # pylint: disable=unused-argument
        """
        Pull out the data into dictionary.
        """
        # Check for presense of required tags in xml.
        if len(xml_object.xpath(cls._child_tag_name)) == 0:
            raise ValueError(
                "Poll_question definition must include \
                at least one 'answer' tag"
            )

        xml_object_copy = deepcopy(xml_object)
        answers = []
        for element_answer in xml_object_copy.findall(cls._child_tag_name):
            answer_id = element_answer.get("id", None)
            if answer_id:
                answers.append({"id": answer_id, "text": stringify_children(element_answer)})
            xml_object_copy.remove(element_answer)

        definition = {"answers": answers, "question": stringify_children(xml_object_copy)}
        children = []
        return (definition, children)

    @classmethod
    def clean_metadata_from_xml(cls, xml_object, excluded_fields=()):
        """
        Remove any attribute named for a field with scope Scope.settings from the supplied
        xml_object
        """

        for field_name, field in getattr(cls, "fields", {}).items():  # pylint: disable=no-member
            if (
                field.scope == Scope.settings
                and field_name not in excluded_fields
                and xml_object.get(field_name) is not None
            ):
                del xml_object.attrib[field_name]

    @staticmethod
    def file_to_xml(file_object):
        """
        Used when this module wants to parse a file object to xml
        that will be converted to the definition.

        Returns an lxml Element
        """
        return etree.parse(file_object, parser=EDX_XML_PARSER).getroot()

    @classmethod
    def load_file(cls, filepath, fs, def_id):  # pylint: disable=invalid-name
        """
        Open the specified file in fs, and call cls.file_to_xml on it,
        returning the lxml object.

        Add details and reraise on error.
        """
        try:
            with fs.open(filepath) as xml_file:
                return cls.file_to_xml(xml_file)
        except Exception as err:
            # Add info about where we are, but keep the traceback
            raise Exception(f"Unable to load file contents at path {filepath} for item {def_id}: {err}") from err

    @classmethod
    def load_definition(cls, xml_object, system, def_id, id_generator):
        """
        Load a block from the specified xml_object.
        """

        # VS[compat]
        # The filename attr should have been removed once all 2012-fall courses imported into the CMS and "inline" OLX
        # format deprecated. This never happened, and `filename` is still used, so we have too keep both formats.
        filename = xml_object.get("filename")
        if filename is None:
            definition_xml = copy.deepcopy(xml_object)
            filepath = ""
            aside_children = []
        else:
            filepath = cls._format_filepath(xml_object.tag, filename)

            definition_xml = cls.load_file(filepath, system.resources_fs, def_id)
            usage_id = id_generator.create_usage(def_id)
            aside_children = system.parse_asides(definition_xml, def_id, usage_id, id_generator)

            # Add the attributes from the pointer node
            definition_xml.attrib.update(xml_object.attrib)

        definition_metadata = cls._get_metadata_from_xml(definition_xml)
        cls.clean_metadata_from_xml(definition_xml)
        definition, children = cls.definition_from_xml(definition_xml, system)
        if definition_metadata:
            definition["definition_metadata"] = definition_metadata
        definition["filename"] = [filepath, filename]

        if aside_children:
            definition["aside_children"] = aside_children

        return definition, children

    @classmethod
    def load_metadata(cls, xml_object):
        """
        Read the metadata attributes from this xml_object.

        Returns a dictionary {key: value}.
        """
        metadata = {"xml_attributes": {}}
        for attr, val in xml_object.attrib.items():

            if attr in cls.metadata_to_strip:
                # don't load these
                continue

            if attr not in getattr(cls, "fields", {}):  # pylint: disable=unsupported-membership-test
                metadata["xml_attributes"][attr] = val
            else:
                metadata[attr] = deserialize_field(cls.fields[attr], val)  # pylint: disable=unsubscriptable-object
        return metadata

    @classmethod
    def apply_policy(cls, metadata, policy):
        """
        Add the keys in policy to metadata, after processing them
        through the attrmap.  Updates the metadata dict in place.
        """
        for attr, value in policy.items():
            if attr not in cls.fields:  # pylint: disable=unsupported-membership-test
                # Store unknown attributes coming from policy.json
                # in such a way that they will export to xml unchanged
                metadata["xml_attributes"][attr] = value
            else:
                metadata[attr] = value

    @classmethod
    def parse_xml(cls, node, runtime, keys):
        """
        Use `node` to construct a new block.
        Returns (XBlock): The newly parsed XBlock
        """

        if keys is None:
            def_id = runtime.id_generator.create_definition(node.tag, node.get("url_name"))
            keys = ScopeIds(None, node.tag, def_id, runtime.id_generator.create_usage(def_id))

        aside_children = []

        if is_pointer_tag(node):
            definition_xml, filepath = cls.load_definition_xml(node, runtime, keys.def_id)
            aside_children = runtime.parse_asides(definition_xml, keys.def_id, keys.usage_id, runtime.id_generator)
        else:
            filepath = None
            definition_xml = node

        definition, children = cls.load_definition(definition_xml, runtime, keys.def_id, runtime.id_generator)

        if is_pointer_tag(node):
            definition["filename"] = [filepath, filepath]

        metadata = cls.load_metadata(definition_xml)

        dmdata = definition.get("definition_metadata", "")
        if dmdata:
            metadata["definition_metadata_raw"] = dmdata
            try:
                metadata.update(json.loads(dmdata))
            except Exception as err:  # pylint: disable=broad-exception-caught
                log.debug("Error in loading metadata %r", dmdata, exc_info=True)
                metadata["definition_metadata_err"] = str(err)

        definition_aside_children = definition.pop("aside_children", None)
        if definition_aside_children:
            aside_children.extend(definition_aside_children)

        field_data = {**metadata, **definition, "children": children}
        field_data["xml_attributes"]["filename"] = definition.get("filename", ["", None])

        if "filename" in field_data:
            del field_data["filename"]

        # The "normal" / new way to set field data:
        xblock = runtime.construct_xblock_from_class(cls, keys)
        for key, value_jsonish in field_data.items():
            if key in cls.fields:  # pylint: disable=unsupported-membership-test
                setattr(xblock, key, cls.fields[key].from_json(value_jsonish))  # pylint: disable=unsubscriptable-object
            elif key == "children":
                xblock.children = value_jsonish
            else:
                log.warning(
                    "Imported %s XBlock does not have field %s found in XML.",
                    xblock.scope_ids.block_type,
                    key,
                )

        if aside_children:
            asides_tags = [x.tag for x in aside_children]
            asides = runtime.get_asides(xblock)
            for asd in asides:
                if asd.scope_ids.block_type in asides_tags:
                    xblock.add_aside(asd)
        return xblock

    @classmethod
    def load_definition_xml(cls, node, runtime, def_id):
        """
        Loads definition_xml stored in a dedicated file
        """
        url_name = node.get("url_name")
        filepath = cls._format_filepath(node.tag, name_to_pathname(url_name))
        definition_xml = cls.load_file(filepath, runtime.resources_fs, def_id)
        return definition_xml, filepath

    @classmethod
    def _format_filepath(cls, category, name):
        return f"{category}/{name}.{cls.filename_extension}"

    def add_xml_to_node(self, node):
        """
        For exporting, set data on `node` from ourselves.
        """
        xml_object = self.definition_to_xml()

        # If xml_object is None, we don't know how to serialize this node, but
        # we shouldn't crash out the whole export for it.
        if xml_object is None:
            return

        for aside in self.runtime.get_asides(self):
            if aside.needs_serialization():
                aside_node = etree.Element("unknown_root", nsmap=XML_NAMESPACES)
                aside.add_xml_to_node(aside_node)
                xml_object.append(aside_node)

        not_to_clean_fields = self.metadata_to_not_to_clean.get(self.category, ())
        self.clean_metadata_from_xml(xml_object, excluded_fields=not_to_clean_fields)

        # Set the tag on both nodes so we get the file path right.
        xml_object.tag = self.category
        node.tag = self.category

        # Add the non-inherited metadata

        for attr in sorted(self.get_explicitly_set_fields_by_scope(Scope.settings)):
            # don't want e.g. data_dir
            if (
                attr not in self.metadata_to_strip
                and attr not in self.metadata_to_export_to_policy
                and attr not in not_to_clean_fields
            ):
                # pylint: disable=unsubscriptable-object
                val = serialize_field(self.fields[attr].to_json(getattr(self, attr)))
                try:
                    xml_object.set(attr, val)
                except Exception:  # pylint: disable=broad-except
                    logging.exception(
                        (
                            "Failed to serialize metadata attribute %s with value %s in module %s. "
                            "This could mean data loss!!!"
                        ),
                        attr,
                        val,
                        self.url_name,
                    )

        for key, value in self.xml_attributes.items():
            if key not in self.metadata_to_strip:
                xml_object.set(key, serialize_field(value))

        node.clear()
        node.tag = xml_object.tag
        node.text = xml_object.text
        node.tail = xml_object.tail
        node.attrib.update(xml_object.attrib)
        node.extend(xml_object)

        # Do not override an existing value for the course.
        if not node.get("url_name"):
            node.set("url_name", self.url_name)

        # Special case for course pointers:
        if self.category == "course":
            # add org and course attributes on the pointer tag
            node.set("org", self.location.org)
            node.set("course", self.location.course)

    def definition_to_xml(self, resource_fs=None):  # pylint: disable=unused-argument
        """Return an xml element representing to this definition."""

        poll_str = HTML("<{tag_name}>{text}</{tag_name}>").format(tag_name=self._tag_name, text=self.question)
        xml_object = etree.fromstring(poll_str)
        xml_object.set("display_name", self.display_name)

        def add_child(xml_obj, answer):  # pylint: disable=unused-argument
            # Escape answer text before adding to xml tree.
            answer_text = str(answer["text"])
            child_str = Text("{tag_begin}{text}{tag_end}").format(
                tag_begin=HTML('<{tag_name} id="{id}">').format(tag_name=self._child_tag_name, id=answer["id"]),
                text=answer_text,
                tag_end=HTML("</{tag_name}>").format(tag_name=self._child_tag_name),
            )
            child_node = etree.fromstring(child_str)
            xml_object.append(child_node)

        for answer in self.answers:
            add_child(xml_object, answer)

        return xml_object



"""TO-DO: Write a description of what this XBlock is."""

from importlib.resources import files

from django.utils import translation
from web_fragments.fragment import Fragment
from xblock.core import XBlock
from xblock.fields import Integer, Scope
from xblock.utils.resources import ResourceLoader

resource_loader = ResourceLoader(__name__)


# This Xblock is just to test the strucutre of xblocks-contrib
@XBlock.needs("i18n")
class DiscussionXBlock(XBlock):
    """
    TO-DO: document what your XBlock does.
    """

    # Fields are defined on the class.  You can access them in your code as
    # self.<fieldname>.

    # TO-DO: delete count, and define your own fields.
    count = Integer(
        default=0,
        scope=Scope.user_state,
        help="A simple counter, to show something happening",
    )

    # Indicates that this XBlock has been extracted from edx-platform.
    is_extracted = True

    def resource_string(self, path):
        """Handy helper for getting resources from our kit."""
        return files(__package__).joinpath(path).read_text(encoding="utf-8")

    # TO-DO: change this view to display your data your own way.
    def student_view(self, context=None):
        """
        Create primary view of the DiscussionXBlock, shown to students when viewing courses.
        """
        if context:
            pass  # TO-DO: do something based on the context.

        frag = Fragment()
        frag.add_content(
            resource_loader.render_django_template(
                "templates/discussion.html",
                {
                    "count": self.count,
                },
                i18n_service=self.runtime.service(self, "i18n"),
            )
        )

        frag.add_css(self.resource_string("static/css/discussion.css"))
        frag.add_javascript(self.resource_string("static/js/src/discussion.js"))
        frag.initialize_js("DiscussionXBlock")
        return frag

    # TO-DO: change this handler to perform your own actions.  You may need more
    # than one handler, or you may not need any handlers at all.
    @XBlock.json_handler
    def increment_count(self, data, suffix=""):
        """
        Increments data. An example handler.
        """
        if suffix:
            pass  # TO-DO: Use the suffix when storing data.
        # Just to show data coming in...
        assert data["hello"] == "world"

        self.count += 1
        return {"count": self.count}

    # TO-DO: change this to create the scenarios you'd like to see in the
    # workbench while developing your XBlock.
    @staticmethod
    def workbench_scenarios():
        """Create canned scenario for display in the workbench."""
        return [
            (
                "DiscussionXBlock",
                """<_discussion_extracted/>
                """,
            ),
            (
                "Multiple DiscussionXBlock",
                """<vertical_demo>
                <_discussion_extracted/>
                <_discussion_extracted/>
                <_discussion_extracted/>
                </vertical_demo>
                """,
            ),
        ]

    @staticmethod
    def get_dummy():
        """
        Generate initial i18n with dummy method.
        """
        return translation.gettext_noop("Dummy")



"""Word cloud is ungraded xblock used by students to
generate and view word cloud.

On the client side we show:
If student does not yet answered - `num_inputs` numbers of text inputs.
If student have answered - words he entered and cloud.
"""
import uuid

from django.utils.translation import gettext_noop as _
from web_fragments.fragment import Fragment
from xblock.core import XBlock
from xblock.fields import Boolean, Dict, Integer, List, Scope, String
from xblock.utils.resources import ResourceLoader
from xblock.utils.studio_editable import StudioEditableXBlockMixin

resource_loader = ResourceLoader(__name__)


def pretty_bool(value):
    """Check value for possible `True` value.

    Using this function we can manage different type of Boolean value
    in xml files.
    """
    bool_dict = [True, "True", "true", "T", "t", "1"]
    return value in bool_dict


@XBlock.needs("i18n")
class WordCloudBlock(StudioEditableXBlockMixin, XBlock):
    """
    Word Cloud XBlock.
    """

    # Indicates that this XBlock has been extracted from edx-platform.
    is_extracted = True
    # "data" class attribute is added to make the following test case pass in edx-platform repo.
    # cms/djangoapps/contentstore/tests/test_contentstore.py::ImportRequiredTestCases::test_empty_data_roundtrip
    data = String(default='', scope=Scope.content)

    editable_fields = ["display_name", "num_inputs", "instructions", "num_top_words", "display_student_percents"]

    display_name = String(
        display_name=_("Display Name"),
        help=_("The display name for this component."),
        scope=Scope.settings,
        default="Word cloud"
    )
    instructions = String(
        display_name=_("Instructions"),
        help=_(
            "Add instructions to help learners understand how to use the word cloud. Clear instructions are "
            "important, especially for learners who have accessibility requirements."),
        scope=Scope.settings,
    )
    num_inputs = Integer(
        display_name=_("Inputs"),
        help=_("The number of text boxes available for learners to add words and sentences."),
        scope=Scope.settings,
        default=5,
        values={"min": 1}
    )
    num_top_words = Integer(
        display_name=_("Maximum Words"),
        help=_("The maximum number of words displayed in the generated word cloud."),
        scope=Scope.settings,
        default=250,
        values={"min": 1}
    )
    display_student_percents = Boolean(
        display_name=_("Show Percents"),
        help=_("Statistics are shown for entered words near that word."),
        scope=Scope.settings,
        default=True
    )

    # Fields for descriptor.
    submitted = Boolean(
        help=_("Whether this learner has posted words to the cloud."),
        scope=Scope.user_state,
        default=False
    )
    student_words = List(
        help=_("Student answer."),
        scope=Scope.user_state,
        default=[]
    )
    all_words = Dict(
        help=_("All possible words from all learners."),
        scope=Scope.user_state_summary
    )
    top_words = Dict(
        help=_("Top num_top_words words for word cloud."),
        scope=Scope.user_state_summary
    )

    @staticmethod
    def workbench_scenarios():
        """Create canned scenario for display in the workbench."""
        return [
            (
                "WordCloudBlock",
                """<_word_cloud_extracted/>
                """,
            ),
            (
                "Multiple WordCloudBlock",
                """<vertical_demo>
                <_word_cloud_extracted/>
                <_word_cloud_extracted/>
                <_word_cloud_extracted/>
                </vertical_demo>
                """,
            ),
        ]

    def student_view(self, context=None):  # pylint: disable=W0613
        """
        Create primary view of the WordCloudXBlock, shown to students when viewing courses.
        """
        frag = Fragment()
        frag.add_content(resource_loader.render_django_template(
            "templates/word_cloud.html", {
                'display_name': self.display_name,
                'instructions': self.instructions,
                'element_class': self.scope_ids.block_type,
                'element_id': uuid.uuid1(0),
                'num_inputs': self.num_inputs,
                'range_num_inputs': range(self.num_inputs),
                'submitted': self.submitted,
            },
            i18n_service=self.runtime.service(self, 'i18n')
        ))
        frag.add_css(resource_loader.load_unicode("static/css/word_cloud.css"))
        frag.add_javascript(resource_loader.load_unicode("static/js/src/word_cloud.js"))
        frag.add_javascript(resource_loader.load_unicode("static/js/src/d3.min.js"))
        frag.add_javascript(resource_loader.load_unicode("static/js/src/d3.layout.cloud.js"))
        frag.add_javascript(resource_loader.load_unicode("static/js/src/html_utils.js"))
        frag.initialize_js('WordCloudBlock')
        return frag

    def good_word(self, word):
        """Convert raw word to suitable word."""
        return word.strip().lower()

    def top_dict(self, dict_obj, amount):
        """Return top words from all words, filtered by number of
        occurences

        :param dict_obj: all words
        :type dict_obj: dict
        :param amount: number of words to be in top dict
        :type amount: int
        :rtype: dict
        """
        return dict(
            sorted(
                list(dict_obj.items()),
                key=lambda x: x[1],
                reverse=True
            )[:amount]
        )

    def get_state(self):
        """Return success json answer for client."""
        if self.submitted:
            total_count = sum(self.all_words.values())
            return {
                'status': 'success',
                'submitted': True,
                'display_student_percents': pretty_bool(
                    self.display_student_percents
                ),
                'student_words': {
                    word: self.all_words[word] for word in self.student_words
                },
                'total_count': total_count,
                'top_words': self.prepare_words(self.top_words, total_count),
            }
        else:
            return {
                'status': 'success',
                'submitted': False,
                'display_student_percents': False,
                'student_words': {},
                'total_count': 0,
                'top_words': {}
            }

    @XBlock.json_handler
    def handle_get_state(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        AJAX handler to get the current state of the XBlock

        Args:
            data: dict having request get parameters

        Returns:
            json string
        """
        return self.get_state()

    @XBlock.json_handler
    def handle_submit_state(self, data, suffix=''):  # pylint: disable=unused-argument
        return self.submit_state(data)

    def submit_state(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        AJAX handler to submit the current state of the XBlock

        Args:
            data: dict having request get parameters

        Returns:
            json string
        """

        if self.submitted:
            return {
                'status': 'fail',
                'error': 'You have already posted your data.'
            }

        # Student words from client.
        # FIXME: we must use raw JSON, not a post data (multipart/form-data)
        raw_student_words = data.get('student_words')
        student_words = [word for word in map(self.good_word, raw_student_words) if word]

        self.student_words = student_words

        # FIXME: fix this, when xblock will support mutable types.
        # Now we use this hack.
        # speed issues
        temp_all_words = self.all_words

        self.submitted = True

        # Save in all_words.
        for word in self.student_words:
            temp_all_words[word] = temp_all_words.get(word, 0) + 1

        # Update top_words.
        self.top_words = self.top_dict(
            temp_all_words,
            self.num_top_words
        )

        # Save all_words in database.
        self.all_words = temp_all_words

        return self.get_state()

    def prepare_words(self, top_words, total_count):
        """Convert words dictionary for client API.

        :param top_words: Top words dictionary
        :type top_words: dict
        :param total_count: Total number of words
        :type total_count: int
        :rtype: list of dicts. Every dict is 3 keys: text - actual word,
            size - counter of word, percent - percent in top_words dataset.

        Calculates corrected percents for every top word:

        For every word except last, it calculates rounded percent.
        For the last is 100 - sum of all other percents.

        """
        list_to_return = []
        percents = 0
        sorted_top_words = sorted(top_words.items(), key=lambda x: x[0].lower())
        for num, word_tuple in enumerate(sorted_top_words):
            if num == len(top_words) - 1:
                percent = 100 - percents
            else:
                percent = round((100.0 * word_tuple[1]) / total_count)
                percents += percent
            list_to_return.append(
                {
                    'text': word_tuple[0],
                    'size': word_tuple[1],
                    'percent': percent
                }
            )
        return list_to_return

    def index_dictionary(self):
        """
        Return dictionary prepared with block content and type for indexing.
        """
        # return key/value fields in a Python dict object
        # values may be numeric / string or dict
        # default implementation is an empty dict

        xblock_body = super().index_dictionary()

        index_body = {
            "display_name": self.display_name,
            "instructions": self.instructions,
        }

        if "content" in xblock_body:
            xblock_body["content"].update(index_body)
        else:
            xblock_body["content"] = index_body

        xblock_body["content_type"] = "Word Cloud"

        return xblock_body



"""
A mixin class for LTI 2.0 functionality.  This is really just done to refactor the code to
keep the LTIBlock class from getting too big
"""


import base64
import hashlib
import json
import logging
import math
import re
from unittest import mock
from urllib import parse

from django.conf import settings
from oauthlib.oauth1 import Client
from webob import Response
from xblock.core import XBlock

log = logging.getLogger(__name__)

LTI_2_0_REST_SUFFIX_PARSER = re.compile(r"^user/(?P<anon_id>\w+)", re.UNICODE)
LTI_2_0_JSON_CONTENT_TYPE = 'application/vnd.ims.lis.v2.result+json'


class LTIError(Exception):
    """Error class for LTIBlock and LTI20BlockMixin"""


class LTI20BlockMixin:
    """
    This class MUST be mixed into LTIBlock.  It does not do anything on its own.  It's just factored
    out for modularity.
    """

    #  LTI 2.0 Result Service Support
    @XBlock.handler
    def lti_2_0_result_rest_handler(self, request, suffix):
        """
        Handler function for LTI 2.0 JSON/REST result service.

        See http://www.imsglobal.org/lti/ltiv2p0/uml/purl.imsglobal.org/vocab/lis/v2/outcomes/Result/service.html
        An example JSON object:

        {
            "@context" : "http://purl.imsglobal.org/ctx/lis/v2/Result",
            "@type" : "Result",
            "resultScore" : 0.83,
            "comment" : "This is exceptional work."

        }
        For PUTs, the content type must be "application/vnd.ims.lis.v2.result+json".
        We use the "suffix" parameter to parse out the user from the end of the URL.  An example endpoint url is
        http://localhost:8000/courses/org/num/run/xblock/i4x:;_;_org;_num;_lti;_GUID/handler_noauth/lti_2_0_result_rest_handler/user/<anon_id>
        so suffix is of the form "user/<anon_id>"
        Failures result in 401, 404, or 500s without any body.  Successes result in 200.  Again see
        http://www.imsglobal.org/lti/ltiv2p0/uml/purl.imsglobal.org/vocab/lis/v2/outcomes/Result/service.html
        (Note: this prevents good debug messages for the client, so we might want to change this, or the spec)

        Arguments:
            request (xblock.django.request.DjangoWebobRequest):  Request object for current HTTP request
            suffix (unicode):  request path after "lti_2_0_result_rest_handler/".  expected to be "user/<anon_id>"

        Returns:
            webob.response:  response to this request.  See above for details.
        """
        if settings.DEBUG:
            self._log_correct_authorization_header(request)

        if not self.accept_grades_past_due and self.is_past_due():
            return Response(status=404)  # have to do 404 due to spec, but 400 is better, with error msg in body

        try:
            anon_id = self.parse_lti_2_0_handler_suffix(suffix)
        except LTIError:
            return Response(status=404)  # 404 because a part of the URL (denoting the anon user id) is invalid
        try:
            self.verify_lti_2_0_result_rest_headers(request, verify_content_type=True)
        except LTIError:
            return Response(status=401)  # Unauthorized in this case.  401 is right

        real_user = self.runtime.service(self, 'user').get_user_by_anonymous_id(anon_id)
        if not real_user:  # that means we can't save to database, as we do not have real user id.
            msg = f"[LTI]: Real user not found against anon_id: {anon_id}"
            log.info(msg)
            return Response(status=404)  # have to do 404 due to spec, but 400 is better, with error msg in body
        if request.method == "PUT":
            return self._lti_2_0_result_put_handler(request, real_user)
        elif request.method == "GET":
            return self._lti_2_0_result_get_handler(request, real_user)
        elif request.method == "DELETE":
            return self._lti_2_0_result_del_handler(request, real_user)
        else:
            return Response(status=404)  # have to do 404 due to spec, but 405 is better, with error msg in body

    def _log_correct_authorization_header(self, request):
        """
        Helper function that logs proper HTTP Authorization header for a given request

        Used only in debug situations, this logs the correct Authorization header based on
        the request header and body according to OAuth 1 Body signing

        Arguments:
            request (xblock.django.request.DjangoWebobRequest):  Request object to log Authorization header for

        Returns:
            nothing
        """
        sha1 = hashlib.sha1()
        sha1.update(request.body)
        oauth_body_hash = str(base64.b64encode(sha1.digest()))
        log.debug(f"[LTI] oauth_body_hash = {oauth_body_hash}")
        client_key, client_secret = self.get_client_key_secret()
        client = Client(client_key, client_secret)
        mock_request = mock.Mock(
            uri=str(parse.unquote(request.url)),
            headers=request.headers,
            body="",
            decoded_body="",
            http_method=str(request.method),
        )
        params = client.get_oauth_params(mock_request)
        mock_request.oauth_params = params
        mock_request.oauth_params.append(('oauth_body_hash', oauth_body_hash))
        sig = client.get_oauth_signature(mock_request)
        mock_request.oauth_params.append(('oauth_signature', sig))

        _, headers, _ = client._render(mock_request)  # pylint: disable=protected-access
        log.debug(
            "\n\n#### COPY AND PASTE AUTHORIZATION HEADER ####\n%s\n####################################\n\n",
            headers['Authorization']
        )

    def parse_lti_2_0_handler_suffix(self, suffix):
        """
        Parser function for HTTP request path suffixes

        parses the suffix argument (the trailing parts of the URL) of the LTI2.0 REST handler.
        must be of the form "user/<anon_id>".  Returns anon_id if match found, otherwise raises LTIError

        Arguments:
            suffix (unicode):  suffix to parse

        Returns:
            unicode: anon_id if match found

        Raises:
            LTIError if suffix cannot be parsed or is not in its expected form
        """
        if suffix:
            match_obj = LTI_2_0_REST_SUFFIX_PARSER.match(suffix)
            if match_obj:
                return match_obj.group('anon_id')
        # fall-through handles all error cases
        msg = "No valid user id found in endpoint URL"
        log.info(f"[LTI]: {msg}")
        raise LTIError(msg)

    def _round_away_from_zero(self, number, digits=0):
        """
        Round numbers using the 'away from zero' strategy as opposed to the
        'Banker's rounding strategy.' The strategy refers to how we round when
        a number is half way between two numbers.  eg. 0.5, 1.5, etc. In python 3
        numbers round towards even. So 0.5 would round to 0 but 1.5 would round to 2.

        See here for more on floating point rounding strategies:
        https://en.wikipedia.org/wiki/IEEE_754#Rounding_rules

        We want to continue to round away from zero so that student grades remain
        consistent and don't suddenly change.
        """
        p = 10.0 ** digits

        if number >= 0:
            return float(math.floor((number * p) + 0.5)) / p
        else:
            return float(math.ceil((number * p) - 0.5)) / p

    def _lti_2_0_result_get_handler(self, _request, real_user):
        """
        Helper request handler for GET requests to LTI 2.0 result endpoint

        GET handler for lti_2_0_result.  Assumes all authorization has been checked.

        Arguments:
            request (xblock.django.request.DjangoWebobRequest):  Request object (unused)
            real_user (django.contrib.auth.models.User):  Actual user linked to anon_id in request path suffix

        Returns:
            webob.response:  response to this request, in JSON format with status 200 if success
        """
        base_json_obj = {
            "@context": "http://purl.imsglobal.org/ctx/lis/v2/Result",
            "@type": "Result"
        }
        self.runtime.service(self, 'rebind_user').rebind_noauth_module_to_user(self, real_user)
        if self.module_score is None:  # In this case, no score has been ever set
            return Response(json.dumps(base_json_obj).encode('utf-8'), content_type=LTI_2_0_JSON_CONTENT_TYPE)

        # Fall through to returning grade and comment
        base_json_obj['resultScore'] = self._round_away_from_zero(self.module_score, 2)
        base_json_obj['comment'] = self.score_comment
        return Response(json.dumps(base_json_obj).encode('utf-8'), content_type=LTI_2_0_JSON_CONTENT_TYPE)

    def _lti_2_0_result_del_handler(self, _request, real_user):
        """
        Helper request handler for DELETE requests to LTI 2.0 result endpoint

        DELETE handler for lti_2_0_result.  Assumes all authorization has been checked.

        Arguments:
            request (xblock.django.request.DjangoWebobRequest):  Request object (unused)
            real_user (django.contrib.auth.models.User):  Actual user linked to anon_id in request path suffix

        Returns:
            webob.response:  response to this request.  status 200 if success
        """
        self.clear_user_module_score(real_user)
        return Response(status=200)

    def _lti_2_0_result_put_handler(self, request, real_user):
        """
        Helper request handler for PUT requests to LTI 2.0 result endpoint

        PUT handler for lti_2_0_result.  Assumes all authorization has been checked.

        Arguments:
            request (xblock.django.request.DjangoWebobRequest):  Request object
            real_user (django.contrib.auth.models.User):  Actual user linked to anon_id in request path suffix

        Returns:
            webob.response:  response to this request.  status 200 if success.  404 if body of PUT request is malformed
        """
        try:
            (score, comment) = self.parse_lti_2_0_result_json(request.body.decode('utf-8'))
        except LTIError:
            return Response(status=404)  # have to do 404 due to spec, but 400 is better, with error msg in body

        # According to http://www.imsglobal.org/lti/ltiv2p0/ltiIMGv2p0.html#_Toc361225514
        # PUTting a JSON object with no "resultScore" field is equivalent to a DELETE.
        if score is None:
            self.clear_user_module_score(real_user)
            return Response(status=200)

        # Fall-through record the score and the comment in the block
        self.set_user_module_score(real_user, score, self.max_score(), comment=comment)
        return Response(status=200)

    def clear_user_module_score(self, user):
        """
        Clears the module user state, including grades and comments, and also scoring in db's courseware_studentmodule

        Arguments:
            user (django.contrib.auth.models.User):  Actual user whose module state is to be cleared

        Returns:
            nothing
        """
        self.set_user_module_score(user, None, None, score_deleted=True)

    def set_user_module_score(self, user, score, max_score, *, comment="", score_deleted=False):
        """
        Sets the module user state, including grades and comments, and also scoring in db's courseware_studentmodule

        Arguments:
            user (django.contrib.auth.models.User):  Actual user whose module state is to be set
            score (float):  user's numeric score to set.  Must be in the range [0.0, 1.0]
            max_score (float):  max score that could have been achieved on this module
            comment (unicode):  comments provided by the grader as feedback to the student

        Returns:
            nothing
        """
        if score is not None and max_score is not None:
            scaled_score = score * max_score
        else:
            scaled_score = None

        self.runtime.service(self, 'rebind_user').rebind_noauth_module_to_user(self, user)

        # have to publish for the progress page...
        self.runtime.publish(
            self,
            'grade',
            {
                'value': scaled_score,
                'max_value': max_score,
                'user_id': user.id,
                'score_deleted': score_deleted,
            },
        )
        self.module_score = scaled_score
        self.score_comment = comment

    def verify_lti_2_0_result_rest_headers(self, request, verify_content_type=True):
        """
        Helper method to validate LTI 2.0 REST result service HTTP headers.  returns if correct, else raises LTIError

        Arguments:
            request (xblock.django.request.DjangoWebobRequest):  Request object
            verify_content_type (bool):  If true, verifies the content type of the request is that spec'ed by LTI 2.0

        Returns:
            nothing, but will only return if verification succeeds

        Raises:
            LTIError if verification fails
        """
        content_type = request.headers.get('Content-Type')
        if verify_content_type and content_type != LTI_2_0_JSON_CONTENT_TYPE:
            log.info(f"[LTI]: v2.0 result service -- bad Content-Type: {content_type}")
            raise LTIError(
                "For LTI 2.0 result service, Content-Type must be {}.  Got {}".format(LTI_2_0_JSON_CONTENT_TYPE,
                                                                                      content_type))
        try:
            self.verify_oauth_body_sign(request, content_type=LTI_2_0_JSON_CONTENT_TYPE)
        except (ValueError, LTIError) as err:
            log.info(f"[LTI]: v2.0 result service -- OAuth body verification failed:  {str(err)}")
            raise LTIError(str(err))  # lint-amnesty, pylint: disable=raise-missing-from

    def parse_lti_2_0_result_json(self, json_str):
        """
        Helper method for verifying LTI 2.0 JSON object contained in the body of the request.

        The json_str must be loadable.  It can either be an dict (object) or an array whose first element is an dict,
        in which case that first dict is considered.
        The dict must have the "@type" key with value equal to "Result",
        "resultScore" key with value equal to a number [0, 1],
        The "@context" key must be present, but we don't do anything with it.  And the "comment" key may be
        present, in which case it must be a string.

        Arguments:
            json_str (unicode):  The body of the LTI 2.0 results service request, which is a JSON string]

        Returns:
            (float, str):  (score, [optional]comment) if verification checks out

        Raises:
            LTIError (with message) if verification fails
        """
        try:
            json_obj = json.loads(json_str)
        except (ValueError, TypeError):
            msg = f"Supplied JSON string in request body could not be decoded: {json_str}"
            log.info(f"[LTI] {msg}")
            raise LTIError(msg)  # lint-amnesty, pylint: disable=raise-missing-from

        # the standard supports a list of objects, who knows why. It must contain at least 1 element, and the
        # first element must be a dict
        if not isinstance(json_obj, dict):
            if isinstance(json_obj, list) and len(json_obj) >= 1 and isinstance(json_obj[0], dict):
                json_obj = json_obj[0]
            else:
                msg = ("Supplied JSON string is a list that does not contain an object as the first element. {}"
                       .format(json_str))
                log.info(f"[LTI] {msg}")
                raise LTIError(msg)

        # '@type' must be "Result"
        result_type = json_obj.get("@type")
        if result_type != "Result":
            msg = f"JSON object does not contain correct @type attribute (should be 'Result', is {result_type})"
            log.info(f"[LTI] {msg}")
            raise LTIError(msg)

        # '@context' must be present as a key
        REQUIRED_KEYS = ["@context"]  # pylint: disable=invalid-name
        for key in REQUIRED_KEYS:
            if key not in json_obj:
                msg = f"JSON object does not contain required key {key}"
                log.info(f"[LTI] {msg}")
                raise LTIError(msg)

        # 'resultScore' is not present.  If this was a PUT this means it's actually a DELETE according
        # to the LTI spec.  We will indicate this by returning None as score, "" as comment.
        # The actual delete will be handled by the caller
        if "resultScore" not in json_obj:
            return None, json_obj.get('comment', "")

        # if present, 'resultScore' must be a number between 0 and 1 inclusive
        try:
            score = float(json_obj.get('resultScore', "unconvertable"))  # Check if float is present and the right type
            if not 0 <= score <= 1:
                msg = 'score value outside the permitted range of 0-1.'
                log.info(f"[LTI] {msg}")
                raise LTIError(msg)
        except (TypeError, ValueError) as err:
            msg = f"Could not convert resultScore to float: {str(err)}"
            log.info(f"[LTI] {msg}")
            raise LTIError(msg)  # lint-amnesty, pylint: disable=raise-missing-from

        return score, json_obj.get('comment', "")



"""
AnnotatableXBlock allows instructors to add interactive annotations to course content.
Annotations can have configurable attributes such as title, body, problem index, and highlight color.
This block enhances the learning experience by enabling students to view embedded comments, questions, or explanations.
The block supports internationalization (i18n) for multilingual courses.
"""

import logging
import textwrap
import uuid

import markupsafe
from django.utils.translation import gettext_noop as _
from lxml import etree
from web_fragments.fragment import Fragment
from xblock.core import XBlock
from xblock.fields import Scope, String, XMLString
from xblock.utils.resources import ResourceLoader

log = logging.getLogger(__name__)

resource_loader = ResourceLoader(__name__)


@XBlock.needs("i18n")
class AnnotatableBlock(XBlock):
    """
    AnnotatableXBlock allows instructors to create annotated content that students can view interactively.
    Annotations can be styled and customized, with internationalization support for multilingual environments.
    """

    # Indicates that this XBlock has been extracted from edx-platform.
    is_extracted = True

    display_name = String(
        display_name=_("Display Name"),
        help=_("The display name for this component."),
        scope=Scope.settings,
        default=_("Annotation"),
    )

    data = XMLString(
        help=_("XML data for the annotation"),
        scope=Scope.content,
        default=textwrap.dedent(
            markupsafe.Markup(
                """
                <annotatable>
                    <instructions>
                        <p>Enter your (optional) instructions for the exercise in HTML format.</p>
                        <p>Annotations are specified by an <code>{}annotation{}</code> tag which may
                        may have the following attributes:</p>
                        <ul class="instructions-template">
                            <li><code>title</code> (optional). Title of the annotation. Defaults to
                            <i>Commentary</i> if omitted.</li>
                            <li><code>body</code> (<b>required</b>). Text of the annotation.</li>
                            <li><code>problem</code> (optional). Numeric index of the problem
                            associated with this annotation. This is a zero-based index, so the first
                            problem on the page would have <code>problem="0"</code>.</li>
                            <li><code>highlight</code> (optional). Possible values: yellow, red,
                            orange, green, blue, or purple. Defaults to yellow if this attribute is
                            omitted.</li>
                        </ul>
                    </instructions>
                    <p>Add your HTML with annotation spans here.</p>
                    <p>Lorem ipsum dolor sit amet, consectetur adipiscing elit.
                    <annotation title="My title" body="My comment" highlight="yellow" problem="0">
                    Ut sodales laoreet est, egestas gravida felis egestas nec.</annotation> Aenean
                    at volutpat erat. Cras commodo viverra nibh in aliquam.</p>
                    <p>Nulla facilisi. <annotation body="Basic annotation example." problem="1">
                    Pellentesque id vestibulum libero.</annotation> Suspendisse potenti. Morbi
                    scelerisque nisi vitae felis dictum mattis. Nam sit amet magna elit. Nullam
                    volutpat cursus est, sit amet sagittis odio vulputate et. Curabitur euismod, orci
                    in vulputate imperdiet, augue lorem tempor purus, id aliquet augue turpis a est.
                    Aenean a sagittis libero. Praesent fringilla pretium magna, non condimentum risus
                    elementum nec. Pellentesque faucibus elementum pharetra. Pellentesque vitae metus
                    eros.</p>
                </annotatable>
                """
            ).format(markupsafe.escape("<"), markupsafe.escape(">"))
        ),
    )

    # List of supported highlight colors for annotations
    HIGHLIGHT_COLORS = ["yellow", "orange", "purple", "blue", "green"]

    def _get_annotation_class_attr(self, index, el):  # pylint: disable=unused-argument
        """Returns a dict with the CSS class attribute to set on the annotation
        and an XML key to delete from the element.
        """

        attr = {}
        cls = ["annotatable-span", "highlight"]
        highlight_key = "highlight"
        color = el.get(highlight_key)

        if color is not None:
            if color in self.HIGHLIGHT_COLORS:
                cls.append("highlight-" + color)
            attr["_delete"] = highlight_key
        attr["value"] = " ".join(cls)

        return {"class": attr}

    def _get_annotation_data_attr(self, index, el):  # pylint: disable=unused-argument
        """Returns a dict in which the keys are the HTML data attributes
        to set on the annotation element. Each data attribute has a
        corresponding 'value' and (optional) '_delete' key to specify
        an XML attribute to delete.
        """

        data_attrs = {}
        attrs_map = {
            "body": "data-comment-body",
            "title": "data-comment-title",
            "problem": "data-problem-id",
        }

        for xml_key, html_key in attrs_map.items():
            if xml_key in el.attrib:
                value = el.get(xml_key, "")
                data_attrs[html_key] = {"value": value, "_delete": xml_key}

        return data_attrs

    def _render_annotation(self, index, el):
        """Renders an annotation element for HTML output."""
        attr = {}
        attr.update(self._get_annotation_class_attr(index, el))
        attr.update(self._get_annotation_data_attr(index, el))

        el.tag = "span"

        for key, value in attr.items():
            el.set(key, value["value"])
            if "_delete" in value and value["_delete"] is not None:
                delete_key = value["_delete"]
                del el.attrib[delete_key]

    def _render_content(self):
        """Renders annotatable content with annotation spans and returns HTML."""

        xmltree = etree.fromstring(self.data)
        self._extract_instructions(xmltree)

        xmltree.tag = "div"
        if "display_name" in xmltree.attrib:
            del xmltree.attrib["display_name"]

        index = 0
        for el in xmltree.findall(".//annotation"):
            self._render_annotation(index, el)
            index += 1

        return etree.tostring(xmltree, encoding="unicode")

    def _extract_instructions(self, xmltree):
        """Removes <instructions> from the xmltree and returns them as a string, otherwise None."""
        instructions = xmltree.find("instructions")
        if instructions is not None:
            instructions.tag = "div"
            xmltree.remove(instructions)
            return etree.tostring(instructions, encoding="unicode")
        return None

    def get_html(self):
        """Returns the HTML representation of the XBlock for student view."""
        return {
            "element_id": uuid.uuid1(0),
            "display_name": self.display_name,
            "instructions_html": self._extract_instructions(etree.fromstring(self.data)),
            "content_html": self._render_content(),
        }

    def student_view(self, context=None):  # pylint: disable=unused-argument
        """Renders the output that a student will see."""
        frag = Fragment()
        frag.add_content(
            resource_loader.render_django_template(
                "templates/annotatable.html",
                self.get_html(),
                i18n_service=self.runtime.service(self, "i18n"),
            )
        )
        frag.add_css(resource_loader.load_unicode("static/css/annotatable.css"))
        frag.add_javascript(resource_loader.load_unicode("static/js/src/annotatable.js"))
        frag.initialize_js("Annotatable")
        return frag

    def studio_view(self, context=None):  # pylint: disable=unused-argument
        """Return the studio view."""
        frag = Fragment()
        frag.add_content(
            resource_loader.render_django_template(
                "templates/annotatable_editor.html",
                {
                    "data": self.data,
                },
                i18n_service=self.runtime.service(self, "i18n"),
            )
        )

        frag.add_css(resource_loader.load_unicode("static/css/annotatable_editor.css"))
        frag.add_javascript(resource_loader.load_unicode("static/js/src/annotatable_editor.js"))
        frag.initialize_js("XMLEditingDescriptor")
        return frag

    @XBlock.json_handler
    def submit_studio_edits(self, data, suffix=""):  # pylint: disable=unused-argument
        """AJAX handler for saving the studio edits."""
        display_name = data.get("display_name")
        xml_data = data.get("data")

        if display_name is not None:
            self.display_name = display_name
        if xml_data is not None:
            self.data = xml_data

        return {"result": "success"}

    @staticmethod
    def workbench_scenarios():
        """Defines scenarios for displaying the XBlock in the XBlock workbench."""
        return [
            ("AnnotatableXBlock", "<_annotatable_extracted/>"),
            (
                "Multiple AnnotatableXBlock",
                """
                <vertical_demo>
                    <_annotatable_extracted/>
                    <_annotatable_extracted/>
                    <_annotatable_extracted/>
                </vertical_demo>
            """,
            ),
        ]



"""
THIS MODULE IS DEPRECATED IN FAVOR OF https://github.com/openedx/xblock-lti-consumer

Learning Tools Interoperability (LTI) module.


Resources
---------

Theoretical background and detailed specifications of LTI can be found on:

    http://www.imsglobal.org/LTI/v1p1p1/ltiIMGv1p1p1.html

This module is based on the version 1.1.1 of the LTI specifications by the
IMS Global authority. For authentication, it uses OAuth1.

When responding back to the LTI tool provider, we must issue a correct
response. Types of responses and their message payload is available at:

    Table A1.2 Interpretation of the 'CodeMajor/severity' matrix.
    http://www.imsglobal.org/gws/gwsv1p0/imsgws_wsdlBindv1p0.html

A resource to test the LTI protocol (PHP realization):

    http://www.imsglobal.org/developers/LTI/test/v1p1/lms.php

We have also begun to add support for LTI 1.2/2.0.  We will keep this
docstring in synch with what support is available.  The first LTI 2.0
feature to be supported is the REST API results service, see specification
at
http://www.imsglobal.org/lti/ltiv2p0/uml/purl.imsglobal.org/vocab/lis/v2/outcomes/Result/service.html

What is supported:
------------------

1.) Display of simple LTI in iframe or a new window.
2.) Multiple LTI components on a single page.
3.) The use of multiple LTI providers per course.
4.) Use of advanced LTI component that provides back a grade.::

    A) LTI 1.1.1 XML endpoint
        a.) The LTI provider sends back a grade to a specified URL.
        b.) Currently only action "update" is supported. "Read", and "delete"
            actions initially weren't required.
    B) LTI 2.0 Result Service JSON REST endpoint
        (http://www.imsglobal.org/lti/ltiv2p0/uml/purl.imsglobal.org/vocab/lis/v2/outcomes/Result/service.html)
        a.) Discovery of all such LTI http endpoints for a course.  External tools GET from this discovery
            endpoint and receive URLs for interacting with individual grading units.
            (see lms/djangoapps/courseware/views/views.py:get_course_lti_endpoints)
        b.) GET, PUT and DELETE in LTI Result JSON binding
            (http://www.imsglobal.org/lti/ltiv2p0/mediatype/application/vnd/ims/lis/v2/result+json/index.html)
            for a provider to synchronize grades into edx-platform.  Reading, Setting, and Deleteing
            Numeric grades between 0 and 1 and text + basic HTML feedback comments are supported, via
            GET / PUT / DELETE HTTP methods respectively
"""

import base64
import datetime
import hashlib
import logging
import textwrap
from unittest import mock
from urllib import parse
from xml.sax.saxutils import escape

import markupsafe
import nh3
import oauthlib.oauth1
from django.conf import settings
from django.utils.translation import gettext_noop as _
from lxml import etree
from oauthlib.oauth1.rfc5849 import signature
from opaque_keys.edx.keys import UsageKey
from pytz import UTC
from web_fragments.fragment import Fragment
from webob import Response
from xblock.core import List, Scope, String, XBlock
from xblock.fields import Boolean, Float

try:
    from xblock.utils.resources import ResourceLoader
    from xblock.utils.studio_editable import StudioEditableXBlockMixin
except ModuleNotFoundError:
    from xblockutils.resources import ResourceLoader
    from xblockutils.studio_editable import StudioEditableXBlockMixin

from .lti_2_util import LTI20BlockMixin, LTIError

# The anonymous user ID for the user in the course.
ATTR_KEY_ANONYMOUS_USER_ID = 'edx-platform.anonymous_user_id'
# The user's role in the course ('staff', 'instructor', or 'student').
ATTR_KEY_USER_ROLE = 'edx-platform.user_role'

resource_loader = ResourceLoader(__name__)

log = logging.getLogger(__name__)

DOCS_ANCHOR_TAG_OPEN = (
    "<a rel='noopener' target='_blank' "
    "href='https://edx.readthedocs.io/projects/edx-partner-course-staff/en/latest/exercises_tools/lti_component.html'>"
)
BREAK_TAG = '<br />'


class LTIFields:
    """
    Fields to define and obtain LTI tool from provider are set here,
    except credentials, which should be set in course settings::

    `lti_id` is id to connect tool with credentials in course settings. It should not contain :: (double semicolon)
    `launch_url` is launch URL of tool.
    `custom_parameters` are additional parameters to navigate to proper book and book page.

    For example, for Vitalsource provider, `launch_url` should be
    *https://bc-staging.vitalsource.com/books/book*,
    and to get to proper book and book page, you should set custom parameters as::

        vbid=put_book_id_here
        book_location=page/put_page_number_here

    Default non-empty URL for `launch_url` is needed due to oauthlib demand (URL scheme should be presented)::

        https://github.com/idan/oauthlib/blob/master/oauthlib/oauth1/rfc5849/signature.py#L136
    """
    display_name = String(
        display_name=_("Display Name"),
        help=_(
            "The display name for this component. "
            "Analytics reports may also use the display name to identify this component."
        ),
        scope=Scope.settings,
        default="LTI",
    )

    lti_id = String(
        display_name=_("LTI ID"),
        help=markupsafe.escape(_(
            "Enter the LTI ID for the external LTI provider.  "
            "This value must be the same LTI ID that you entered in the "
            "LTI Passports setting on the Advanced Settings page."
            "{break_tag}See {docs_anchor_open}the edX LTI documentation{anchor_close} for more details on this setting."
        )).format(
            break_tag=markupsafe.Markup(BREAK_TAG),
            docs_anchor_open=markupsafe.Markup(DOCS_ANCHOR_TAG_OPEN),
            anchor_close=markupsafe.Markup("</a>")
        ),
        default='',
        scope=Scope.settings
    )

    launch_url = String(
        display_name=_("LTI URL"),
        help=markupsafe.escape(_(
            "Enter the URL of the external tool that this component launches. "
            "This setting is only used when Hide External Tool is set to False."
            "{break_tag}See {docs_anchor_open}the edX LTI documentation{anchor_close} for more details on this setting."
        )).format(
            break_tag=markupsafe.Markup(BREAK_TAG),
            docs_anchor_open=markupsafe.Markup(DOCS_ANCHOR_TAG_OPEN),
            anchor_close=markupsafe.Markup("</a>")
        ),
        default='http://www.example.com',
        scope=Scope.settings
    )

    custom_parameters = List(
        display_name=_("Custom Parameters"),
        help=markupsafe.escape(_(
            "Add the key/value pair for any custom parameters, such as the page your e-book should open to or "
            "the background color for this component."
            "{break_tag}See {docs_anchor_open}the edX LTI documentation{anchor_close} for more details on this setting."
        )).format(
            break_tag=markupsafe.Markup(BREAK_TAG),
            docs_anchor_open=markupsafe.Markup(DOCS_ANCHOR_TAG_OPEN),
            anchor_close=markupsafe.Markup("</a>")
        ),
        scope=Scope.settings
    )

    open_in_a_new_page = Boolean(
        display_name=_("Open in New Page"),
        help=_(
            "Select True if you want students to click a link that opens the LTI tool in a new window. "
            "Select False if you want the LTI content to open in an IFrame in the current page. "
            "This setting is only used when Hide External Tool is set to False.  "
        ),
        default=True,
        scope=Scope.settings
    )

    has_score = Boolean(
        display_name=_("Scored"),
        help=_(
            "Select True if this component will receive a numerical score from the external LTI system."
        ),
        default=False,
        scope=Scope.settings
    )

    weight = Float(
        display_name=_("Weight"),
        help=_(
            "Enter the number of points possible for this component.  "
            "The default value is 1.0.  "
            "This setting is only used when Scored is set to True."
        ),
        default=1.0,
        scope=Scope.settings,
        values={"min": 0},
    )

    module_score = Float(
        help=_("The score kept in the xblock KVS -- duplicate of the published score in django DB"),
        default=None,
        scope=Scope.user_state
    )

    score_comment = String(
        help=_("Comment as returned from grader, LTI2.0 spec"),
        default="",
        scope=Scope.user_state
    )

    hide_launch = Boolean(
        display_name=_("Hide External Tool"),
        help=_(
            "Select True if you want to use this component as a placeholder for syncing with an external grading  "
            "system rather than launch an external tool.  "
            "This setting hides the Launch button and any IFrames for this component."
        ),
        default=False,
        scope=Scope.settings
    )

    # Users will be presented with a message indicating that their e-mail/username would be sent to a third
    # party application. When "Open in New Page" is not selected, the tool automatically appears without any
    #  user action.
    ask_to_send_username = Boolean(
        display_name=_("Request user's username"),
        # Translators: This is used to request the user's username for a third party service.
        help=_("Select True to request the user's username."),
        default=False,
        scope=Scope.settings
    )

    ask_to_send_email = Boolean(
        display_name=_("Request user's email"),
        # Translators: This is used to request the user's email for a third party service.
        help=_("Select True to request the user's email address."),
        default=False,
        scope=Scope.settings
    )

    description = String(
        display_name=_("LTI Application Information"),
        help=_(
            "Enter a description of the third party application. If requesting username and/or email, use this text"
            " box to inform users why their username and/or email will be forwarded to a third party application."
        ),
        default="",
        scope=Scope.settings
    )

    button_text = String(
        display_name=_("Button Text"),
        help=_(
            "Enter the text on the button used to launch the third party application."
        ),
        default="",
        scope=Scope.settings
    )

    accept_grades_past_due = Boolean(
        display_name=_("Accept grades past deadline"),
        help=_("Select True to allow third party systems to post grades past the deadline."),
        default=True,
        scope=Scope.settings
    )

    editable_fields = (
        "accept_grades_past_due", "ask_to_send_email", "ask_to_send_username", "button_text",
        "custom_parameters", "description", "display_name", "has_score", "hide_launch",
        "launch_url", "lti_id", "open_in_a_new_page", "weight",
    )


@XBlock.needs("i18n")
@XBlock.needs("user")
@XBlock.needs("rebind_user")
class LTIBlock(
    LTIFields,
    LTI20BlockMixin,
    StudioEditableXBlockMixin,
    XBlock,
):
    """
    THIS MODULE IS DEPRECATED IN FAVOR OF https://github.com/openedx/xblock-lti-consumer

    Module provides LTI integration to course.

    Except usual Xmodule structure it proceeds with OAuth signing.
    How it works::

        1. Get credentials from course settings.

        2.  There is minimal set of parameters need to be signed (presented for Vitalsource)::

                user_id
                oauth_callback
                lis_outcome_service_url
                lis_result_sourcedid
                launch_presentation_return_url
                lti_message_type
                lti_version
                roles
                *+ all custom parameters*

            These parameters should be encoded and signed by *OAuth1* together with
            `launch_url` and *POST* request type.

        3. Signing proceeds with client key/secret pair obtained from course settings.
            That pair should be obtained from LTI provider and set into course settings by course author.
            After that signature and other OAuth data are generated.

            OAuth data which is generated after signing is usual::

                oauth_callback
                oauth_nonce
                oauth_consumer_key
                oauth_signature_method
                oauth_timestamp
                oauth_version


        4. All that data is passed to form and sent to LTI provider server by browser via
            autosubmit via JavaScript.

            Form example::

                <form
                    action="${launch_url}"
                    name="ltiLaunchForm-${element_id}"
                    class="ltiLaunchForm"
                    method="post"
                    target="ltiLaunchFrame-${element_id}"
                    encType="application/x-www-form-urlencoded"
                >
                    <input name="launch_presentation_return_url" value="" />
                    <input name="lis_outcome_service_url" value="" />
                    <input name="lis_result_sourcedid" value="" />
                    <input name="lti_message_type" value="basic-lti-launch-request" />
                    <input name="lti_version" value="LTI-1p0" />
                    <input name="oauth_callback" value="about:blank" />
                    <input name="oauth_consumer_key" value="${oauth_consumer_key}" />
                    <input name="oauth_nonce" value="${oauth_nonce}" />
                    <input name="oauth_signature_method" value="HMAC-SHA1" />
                    <input name="oauth_timestamp" value="${oauth_timestamp}" />
                    <input name="oauth_version" value="1.0" />
                    <input name="user_id" value="${user_id}" />
                    <input name="role" value="student" />
                    <input name="oauth_signature" value="${oauth_signature}" />

                    <input name="custom_1" value="${custom_param_1_value}" />
                    <input name="custom_2" value="${custom_param_2_value}" />
                    <input name="custom_..." value="${custom_param_..._value}" />

                    <input type="submit" value="Press to Launch" />
                </form>

        5. LTI provider has same secret key and it signs data string via *OAuth1* and compares signatures.

            If signatures are correct, LTI provider redirects iframe source to LTI tool web page,
            and LTI tool is rendered to iframe inside course.

            Otherwise error message from LTI provider is generated.
    """

    # Indicates that this XBlock has been extracted from edx-platform.
    is_extracted = True

    @property
    def location(self):
        return self.scope_ids.usage_id

    @location.setter
    def location(self, value):
        assert isinstance(value, UsageKey)
        self.scope_ids = self.scope_ids._replace(
            def_id=value,  # Note: assigning a UsageKey as def_id is OK in old mongo / import system but wrong in split
            usage_id=value,
        )

    def max_score(self):
        return self.weight if self.has_score else None

    def get_input_fields(self):  # lint-amnesty, pylint: disable=missing-function-docstring
        # LTI provides a list of default parameters that might be passed as
        # part of the POST data. These parameters should not be prefixed.
        # Likewise, The creator of an LTI link can add custom key/value parameters
        # to a launch which are to be included with the launch of the LTI link.
        # In this case, we will automatically add `custom_` prefix before this parameters.
        # See http://www.imsglobal.org/LTI/v1p1p1/ltiIMGv1p1p1.html#_Toc316828520
        PARAMETERS = [
            "lti_message_type",
            "lti_version",
            "resource_link_title",
            "resource_link_description",
            "user_image",
            "lis_person_name_given",
            "lis_person_name_family",
            "lis_person_name_full",
            "lis_person_contact_email_primary",
            "lis_person_sourcedid",
            "role_scope_mentor",
            "context_type",
            "context_title",
            "context_label",
            "launch_presentation_locale",
            "launch_presentation_document_target",
            "launch_presentation_css_url",
            "launch_presentation_width",
            "launch_presentation_height",
            "launch_presentation_return_url",
            "tool_consumer_info_product_family_code",
            "tool_consumer_info_version",
            "tool_consumer_instance_guid",
            "tool_consumer_instance_name",
            "tool_consumer_instance_description",
            "tool_consumer_instance_url",
            "tool_consumer_instance_contact_email",
        ]

        client_key, client_secret = self.get_client_key_secret()

        # parsing custom parameters to dict
        custom_parameters = {}

        for custom_parameter in self.custom_parameters:
            try:
                param_name, param_value = [p.strip() for p in custom_parameter.split('=', 1)]
            except ValueError:
                _ = self.runtime.service(self, "i18n").ugettext
                msg = _('Could not parse custom parameter: {custom_parameter}. Should be "x=y" string.').format(
                    custom_parameter=f"{custom_parameter!r}"
                )
                raise LTIError(msg)  # lint-amnesty, pylint: disable=raise-missing-from

            # LTI specs: 'custom_' should be prepended before each custom parameter, as pointed in link above.
            if param_name not in PARAMETERS:
                param_name = 'custom_' + param_name

            custom_parameters[str(param_name)] = str(param_value)

        return self.oauth_params(
            custom_parameters,
            client_key,
            client_secret,
        )

    def get_context(self):
        """
        Returns a context.
        """
        # nh3 defaults for
        # ALLOWED_TAGS are
        # {
        #   'a', 'abbr', 'acronym', 'area', 'article', 'aside', 'b', 'bdi', 'bdo',
        #   'blockquote', 'br', 'caption', 'center', 'cite', 'code', 'col', 'colgroup',
        #   'data', 'dd', 'del', 'details', 'dfn', 'div', 'dl', 'dt', 'em', 'figcaption',
        #   'figure', 'footer', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'header', 'hgroup',
        #   'hr', 'i', 'img', 'ins', 'kbd', 'li', 'map', 'mark', 'nav', 'ol', 'p', 'pre',
        #   'q', 'rp', 'rt', 'rtc', 'ruby', 's', 'samp', 'small', 'span', 'strike',
        #   'strong', 'sub', 'summary', 'sup', 'table', 'tbody', 'td', 'th', 'thead',
        #   'time', 'tr', 'tt', 'u', 'ul', 'var', 'wbr'
        # }
        #
        # ALLOWED_ATTRIBUTES are
        # {
        #   'a': {'href', 'hreflang'},
        #   'bdo': {'dir'},
        #   'blockquote': {'cite'},
        #   'col': {'charoff', 'char', 'align', 'span'},
        #   'colgroup': {'align', 'char', 'charoff', 'span'},
        #   'del': {'datetime', 'cite'},
        #   'hr': {'width', 'align', 'size'},
        #   'img': {'height', 'src', 'width', 'alt', 'align'},
        #   'ins': {'datetime', 'cite'},
        #   'ol': {'start'},
        #   'q': {'cite'},
        #   'table': {'align', 'char', 'charoff', 'summary'},
        #   'tbody': {'align', 'char', 'charoff'},
        #   'td': {'rowspan', 'headers', 'charoff', 'colspan', 'char', 'align'},
        #   'tfoot': {'align', 'char', 'charoff'},
        #   'th': {'rowspan', 'headers', 'charoff', 'colspan', 'scope', 'char', 'align'},
        #   'thead': {'charoff', 'char', 'align'},
        #   'tr': {'align', 'char', 'charoff'}
        # }
        #
        # This lets all plaintext through.
        sanitized_comment = nh3.clean(self.score_comment)

        return {
            'input_fields': self.get_input_fields(),

            # These parameters do not participate in OAuth signing.
            'launch_url': self.launch_url.strip(),
            'element_id': self.location.html_id(),
            'element_class': self.scope_ids.block_type,
            'open_in_a_new_page': self.open_in_a_new_page,
            'display_name': self.display_name,
            'form_url': self.runtime.handler_url(self, 'preview_handler').rstrip('/?'),
            'hide_launch': self.hide_launch,
            'has_score': self.has_score,
            'weight': self.weight,
            'module_score': self.module_score,
            'comment': sanitized_comment,
            'description': self.description,
            'ask_to_send_username': self.ask_to_send_username,
            'ask_to_send_email': self.ask_to_send_email,
            'button_text': self.button_text,
            'accept_grades_past_due': self.accept_grades_past_due,
        }

    def student_view(self, _context=None):
        """
        Create primary view of the LTIBlock, shown to students when viewing courses.
        """
        frag = Fragment()
        frag.add_content(
            resource_loader.render_django_template(
                "templates/lti.html", self.get_context(),
                i18n_service=self.runtime.service(self, 'i18n')
            )
        )

        frag.add_css(resource_loader.load_unicode("static/css/lti.css"))
        frag.add_javascript(resource_loader.load_unicode("static/js/src/lti.js"))
        frag.initialize_js("LTIBlock")
        return frag

    @XBlock.handler
    def preview_handler(self, _, __):
        """
        This is called to get context with new oauth params to iframe.
        """
        template = resource_loader.render_django_template("templates/lti_form.html", self.get_context())
        return Response(template, content_type='text/html')

    @XBlock.handler
    def grade_handler(self, request, suffix):  # lint-amnesty, pylint: disable=unused-argument
        """
        This is called by courseware.block_render, to handle an AJAX call.

        Used only for grading. Returns XML response.

        Example of request body from LTI provider::

            <?xml version = "1.0" encoding = "UTF-8"?>
                <imsx_POXEnvelopeRequest xmlns = "some_link (may be not required)">
                    <imsx_POXHeader>
                        <imsx_POXRequestHeaderInfo>
                        <imsx_version>V1.0</imsx_version>
                        <imsx_messageIdentifier>528243ba5241b</imsx_messageIdentifier>
                        </imsx_POXRequestHeaderInfo>
                    </imsx_POXHeader>
                    <imsx_POXBody>
                        <replaceResultRequest>
                        <resultRecord>
                            <sourcedGUID>
                            <sourcedId>feb-123-456-2929::28883</sourcedId>
                            </sourcedGUID>
                            <result>
                            <resultScore>
                                <language>en-us</language>
                                <textString>0.4</textString>
                            </resultScore>
                            </result>
                        </resultRecord>
                        </replaceResultRequest>
                    </imsx_POXBody>
                </imsx_POXEnvelopeRequest>

            Example of correct/incorrect answer XML body:: see response_xml_template.
        """
        response_xml_template = textwrap.dedent("""\
            <?xml version="1.0" encoding="UTF-8"?>
            <imsx_POXEnvelopeResponse xmlns = "http://www.imsglobal.org/services/ltiv1p1/xsd/imsoms_v1p0">
                <imsx_POXHeader>
                    <imsx_POXResponseHeaderInfo>
                        <imsx_version>V1.0</imsx_version>
                        <imsx_messageIdentifier>{imsx_messageIdentifier}</imsx_messageIdentifier>
                        <imsx_statusInfo>
                            <imsx_codeMajor>{imsx_codeMajor}</imsx_codeMajor>
                            <imsx_severity>status</imsx_severity>
                            <imsx_description>{imsx_description}</imsx_description>
                            <imsx_messageRefIdentifier>
                            </imsx_messageRefIdentifier>
                        </imsx_statusInfo>
                    </imsx_POXResponseHeaderInfo>
                </imsx_POXHeader>
                <imsx_POXBody>{response}</imsx_POXBody>
            </imsx_POXEnvelopeResponse>
        """)
        # Returns when `action` is unsupported.
        # Supported actions:
        #   - replaceResultRequest.
        unsupported_values = {
            'imsx_codeMajor': 'unsupported',
            'imsx_description': 'Target does not support the requested operation.',
            'imsx_messageIdentifier': 'unknown',
            'response': ''
        }
        # Returns if:
        #   - past due grades are not accepted and grade is past due
        #   - score is out of range
        #   - can't parse response from TP;
        #   - can't verify OAuth signing or OAuth signing is incorrect.
        failure_values = {
            'imsx_codeMajor': 'failure',
            'imsx_description': 'The request has failed.',
            'imsx_messageIdentifier': 'unknown',
            'response': ''
        }

        if not self.accept_grades_past_due and self.is_past_due():
            failure_values['imsx_description'] = "Grade is past due"
            return Response(response_xml_template.format(**failure_values), content_type="application/xml")

        try:
            imsx_messageIdentifier, sourcedId, score, action = self.parse_grade_xml_body(request.body)
        except Exception as e:  # lint-amnesty, pylint: disable=broad-except
            error_message = "Request body XML parsing error: " + escape(str(e))
            log.debug("[LTI]: " + error_message)  # lint-amnesty, pylint: disable=logging-not-lazy
            failure_values['imsx_description'] = error_message
            return Response(response_xml_template.format(**failure_values), content_type="application/xml")

        # Verify OAuth signing.
        try:
            self.verify_oauth_body_sign(request)
        except (ValueError, LTIError) as e:
            failure_values['imsx_messageIdentifier'] = escape(imsx_messageIdentifier)
            error_message = "OAuth verification error: " + escape(str(e))
            failure_values['imsx_description'] = error_message
            log.debug("[LTI]: " + error_message)  # lint-amnesty, pylint: disable=logging-not-lazy
            return Response(response_xml_template.format(**failure_values), content_type="application/xml")

        real_user = self.runtime.service(self, 'user').get_user_by_anonymous_id(parse.unquote(sourcedId.split(':')[-1]))
        if not real_user:  # that means we can't save to database, as we do not have real user id.
            failure_values['imsx_messageIdentifier'] = escape(imsx_messageIdentifier)
            failure_values['imsx_description'] = "User not found."
            return Response(response_xml_template.format(**failure_values), content_type="application/xml")

        if action == 'replaceResultRequest':
            self.set_user_module_score(real_user, score, self.max_score())

            values = {
                'imsx_codeMajor': 'success',
                'imsx_description': f'Score for {sourcedId} is now {score}',
                'imsx_messageIdentifier': escape(imsx_messageIdentifier),
                'response': '<replaceResultResponse/>'
            }
            log.debug("[LTI]: Grade is saved.")
            return Response(response_xml_template.format(**values), content_type="application/xml")

        unsupported_values['imsx_messageIdentifier'] = escape(imsx_messageIdentifier)
        log.debug("[LTI]: Incorrect action.")
        return Response(response_xml_template.format(**unsupported_values), content_type='application/xml')

    @staticmethod
    def workbench_scenarios():
        """Create canned scenario for display in the workbench."""
        return [
            (
                "LTIBlock",
                """<_lti_extracted/>
                """,
            ),
            (
                "Multiple LTIBlock",
                """<vertical_demo>
                <_lti_extracted/>
                <_lti_extracted/>
                <_lti_extracted/>
                </vertical_demo>
                """,
            ),
        ]

    def get_user_id(self):
        """
        Returns the current user ID, URL-escaped so it is safe to use as a URL component.
        """
        user_id = self.runtime.service(self, 'user').get_current_user().opt_attrs.get(ATTR_KEY_ANONYMOUS_USER_ID)
        assert user_id is not None
        return str(parse.quote(user_id))

    def get_outcome_service_url(self, service_name="grade_handler"):
        """
        Return URL for storing grades.

        To test LTI on sandbox we must use http scheme.

        While testing locally and on Jenkins, mock_lti_server use http.referer
        to obtain scheme, so it is ok to have http(s) anyway.

        The scheme logic is handled in lms/lib/xblock/runtime.py
        """
        return self.runtime.handler_url(self, service_name, thirdparty=True).rstrip('/?')

    def get_resource_link_id(self):
        """
        This is an opaque unique identifier that the TC guarantees will be unique
        within the TC for every placement of the link.

        If the tool / activity is placed multiple times in the same context,
        each of those placements will be distinct.

        This value will also change if the item is exported from one system or
        context and imported into another system or context.

        This parameter is required.

        Example:  u'edx.org-i4x-2-3-lti-31de800015cf4afb973356dbe81496df'

        Hostname, edx.org,
        makes resource_link_id change on import to another system.

        Last part of location, location.name - 31de800015cf4afb973356dbe81496df,
        is random hash, updated by course_id,
        this makes resource_link_id unique inside single course.

        First part of location is tag-org-course-category, i4x-2-3-lti.

        Location.name itself does not change on import to another course,
        but org and course_id change.

        So together with org and course_id in a form of
        i4x-2-3-lti-31de800015cf4afb973356dbe81496df this part of resource_link_id:
        makes resource_link_id to be unique among courses inside same system.
        """
        return str(parse.quote(f"{settings.LMS_BASE}-{self.location.html_id()}"))

    def get_lis_result_sourcedid(self):
        """
        This field contains an identifier that indicates the LIS Result Identifier (if any)
        associated with this launch.  This field identifies a unique row and column within the
        TC gradebook.  This field is unique for every combination of context_id / resource_link_id / user_id.
        This value may change for a particular resource_link_id / user_id  from one launch to the next.
        The TP should only retain the most recent value for this field for a particular resource_link_id / user_id.
        This field is generally optional, but is required for grading.
        """
        return "{context}:{resource_link}:{user_id}".format(
            context=parse.quote(self.context_id),
            resource_link=self.get_resource_link_id(),
            user_id=self.get_user_id()
        )

    def get_course(self):
        """
        Return course by course id.
        """
        return self.runtime.modulestore.get_course(self.location.course_key)

    @property
    def context_id(self):
        """
        Return context_id.

        context_id is an opaque identifier that uniquely identifies the context (e.g., a course)
        that contains the link being launched.
        """
        return str(self.location.course_key)

    @property
    def role(self):
        """
        Get system user role and convert it to LTI role.
        """
        roles = {
            'student': 'Student',
            'staff': 'Administrator',
            'instructor': 'Instructor',
        }
        user_role = self.runtime.service(self, 'user').get_current_user().opt_attrs.get(ATTR_KEY_USER_ROLE)
        return roles.get(user_role, 'Student')

    def get_icon_class(self):
        """ Returns the icon class """
        if self.graded and self.has_score:
            return 'problem'
        return 'other'

    def oauth_params(self, custom_parameters, client_key, client_secret):
        """
        Signs request and returns signature and OAuth parameters.

        `custom_paramters` is dict of parsed `custom_parameter` field
        `client_key` and `client_secret` are LTI tool credentials.

        Also *anonymous student id* is passed to template and therefore to LTI provider.
        """

        client = oauthlib.oauth1.Client(
            client_key=str(client_key),
            client_secret=str(client_secret)
        )

        # Must have parameters for correct signing from LTI:
        body = {
            'user_id': self.get_user_id(),
            'oauth_callback': 'about:blank',
            'launch_presentation_return_url': '',
            'lti_message_type': 'basic-lti-launch-request',
            'lti_version': 'LTI-1p0',
            'roles': self.role,

            # Parameters required for grading:
            'resource_link_id': self.get_resource_link_id(),
            'lis_result_sourcedid': self.get_lis_result_sourcedid(),

            'context_id': self.context_id,
        }

        if self.has_score:
            body.update({
                'lis_outcome_service_url': self.get_outcome_service_url()
            })

        self.user_email = ""  # lint-amnesty, pylint: disable=attribute-defined-outside-init
        self.user_username = ""  # lint-amnesty, pylint: disable=attribute-defined-outside-init

        # Username and email can't be sent in studio mode, because the user object is not defined.
        # To test functionality test in LMS

        real_user_object = self.runtime.service(self, 'user').get_user_by_anonymous_id()
        try:
            self.user_email = real_user_object.email  # lint-amnesty, pylint: disable=attribute-defined-outside-init
        except AttributeError:
            self.user_email = ""  # lint-amnesty, pylint: disable=attribute-defined-outside-init
        try:
            self.user_username = real_user_object.username  # pylint: disable=attribute-defined-outside-init
        except AttributeError:
            self.user_username = ""  # lint-amnesty, pylint: disable=attribute-defined-outside-init

        if self.ask_to_send_username and self.user_username:
            body["lis_person_sourcedid"] = self.user_username
        if self.ask_to_send_email and self.user_email:
            body["lis_person_contact_email_primary"] = self.user_email

        # Appending custom parameter for signing.
        body.update(custom_parameters)

        headers = {
            # This is needed for body encoding:
            'Content-Type': 'application/x-www-form-urlencoded',
        }

        try:
            __, headers, __ = client.sign(
                str(self.launch_url.strip()),
                http_method='POST',
                body=body,
                headers=headers)
        except ValueError:  # Scheme not in url.
            # https://github.com/idan/oauthlib/blob/master/oauthlib/oauth1/rfc5849/signature.py#L136
            # Stubbing headers for now:
            log.info(
                "LTI block %s in course %s does not have oauth parameters correctly configured.",
                self.location,
                self.location.course_key,
            )
            headers = {
                'Content-Type': 'application/x-www-form-urlencoded',
                'Authorization': 'OAuth oauth_nonce="80966668944732164491378916897", \
oauth_timestamp="1378916897", oauth_version="1.0", oauth_signature_method="HMAC-SHA1", \
oauth_consumer_key="", oauth_signature="frVp4JuvT1mVXlxktiAUjQ7%2F1cw%3D"'}

        params = headers['Authorization']
        # Parse headers to pass to template as part of context:
        params = dict([param.strip().replace('"', '').split('=') for param in params.split(',')])

        params['oauth_nonce'] = params['OAuth oauth_nonce']
        del params['OAuth oauth_nonce']

        # oauthlib encodes signature with
        # 'Content-Type': 'application/x-www-form-urlencoded'
        # so '='' becomes '%3D'.
        # We send form via browser, so browser will encode it again,
        # So we need to decode signature back:
        params['oauth_signature'] = parse.unquote(params['oauth_signature']).encode('utf-8').decode('utf8')

        # Add LTI parameters to OAuth parameters for sending in form.
        params.update(body)
        return params

    @classmethod
    def parse_grade_xml_body(cls, body):
        """
        Parses XML from request.body and returns parsed data

        XML body should contain nsmap with namespace, that is specified in LTI specs.

        Returns tuple: imsx_messageIdentifier, sourcedId, score, action

        Raises Exception if can't parse.
        """
        lti_spec_namespace = "http://www.imsglobal.org/services/ltiv1p1/xsd/imsoms_v1p0"
        namespaces = {'def': lti_spec_namespace}

        data = body.strip()
        parser = etree.XMLParser(ns_clean=True, recover=True, encoding='utf-8')
        root = etree.fromstring(data, parser=parser)

        imsx_messageIdentifier = root.xpath("//def:imsx_messageIdentifier", namespaces=namespaces)[0].text or ''
        sourcedId = root.xpath("//def:sourcedId", namespaces=namespaces)[0].text
        score = root.xpath("//def:textString", namespaces=namespaces)[0].text
        action = root.xpath(
            "//def:imsx_POXBody", namespaces=namespaces
        )[0].getchildren()[0].tag.replace('{' + lti_spec_namespace + '}', '')
        # Raise exception if score is not float or not in range 0.0-1.0 regarding spec.
        score = float(score)
        if not 0 <= score <= 1:
            raise LTIError('score value outside the permitted range of 0-1.')

        return imsx_messageIdentifier, sourcedId, score, action

    def verify_oauth_body_sign(self, request, content_type='application/x-www-form-urlencoded'):
        """
        Verify grade request from LTI provider using OAuth body signing.

        Uses http://oauth.googlecode.com/svn/spec/ext/body_hash/1.0/oauth-bodyhash.html::

            This specification extends the OAuth signature to include integrity checks on HTTP request bodies
            with content types other than application/x-www-form-urlencoded.

        Arguments:
            request: DjangoWebobRequest.

        Raises:
            LTIError if request is incorrect.
        """

        client_key, client_secret = self.get_client_key_secret()  # lint-amnesty, pylint: disable=unused-variable
        headers = {
            'Authorization': str(request.headers.get('Authorization')),
            'Content-Type': content_type,
        }

        sha1 = hashlib.sha1()
        sha1.update(request.body)
        oauth_body_hash = base64.b64encode(sha1.digest()).decode('utf-8')
        oauth_params = signature.collect_parameters(headers=headers, exclude_oauth_signature=False)
        oauth_headers = dict(oauth_params)
        oauth_signature = oauth_headers.pop('oauth_signature')
        mock_request_lti_1 = mock.Mock(
            uri=str(parse.unquote(self.get_outcome_service_url())),
            http_method=str(request.method),
            params=list(oauth_headers.items()),
            signature=oauth_signature
        )
        mock_request_lti_2 = mock.Mock(
            uri=str(parse.unquote(request.url)),
            http_method=str(request.method),
            params=list(oauth_headers.items()),
            signature=oauth_signature
        )
        if oauth_body_hash != oauth_headers.get('oauth_body_hash'):
            log.error(
                "OAuth body hash verification failed, provided: %s, calculated: %s, for url: %s, body is: %s",
                oauth_headers.get('oauth_body_hash'),
                oauth_body_hash,
                self.get_outcome_service_url(),
                request.body
            )
            raise LTIError("OAuth body hash verification is failed.")

        if (not signature.verify_hmac_sha1(mock_request_lti_1, client_secret) and not
                signature.verify_hmac_sha1(mock_request_lti_2, client_secret)):
            log.error(
                "OAuth signature verification failed, for headers: %s url: %s method: %s",
                oauth_headers, self.get_outcome_service_url(), str(request.method)
            )
            raise LTIError("OAuth signature verification has failed.")

    def get_client_key_secret(self):
        """
        Obtains client_key and client_secret credentials from current course.
        """
        course = self.get_course()
        for lti_passport in course.lti_passports:
            try:
                lti_id, key, secret = [i.strip() for i in lti_passport.split(':')]
            except ValueError:
                _ = self.runtime.service(self, "i18n").ugettext
                msg = _('Could not parse LTI passport: {lti_passport}. Should be "id:key:secret" string.').format(
                    lti_passport=f'{lti_passport!r}'
                )
                raise LTIError(msg)  # lint-amnesty, pylint: disable=raise-missing-from

            if lti_id == self.lti_id.strip():
                return key, secret
        return '', ''

    def is_past_due(self):
        """
        Is it now past this problem's due date, including grace period?
        """
        due_date = self.due
        if self.graceperiod is not None and due_date:
            close_date = due_date + self.graceperiod
        else:
            close_date = due_date
        return close_date is not None and datetime.datetime.now(UTC) > close_date



"""
HTML XBlock module for displaying raw HTML content.
This XBlock allows users to embed HTML content inside courses.
"""

import copy
import logging
import os
import re
import sys
import uuid
from html.parser import HTMLParser

from django.conf import settings
from django.utils.translation import gettext_noop as _
from fs.errors import ResourceNotFound
from lxml import etree
from opaque_keys.edx.keys import CourseKey, UsageKey
from opaque_keys.edx.locator import LibraryLocatorV2
from path import Path as path
from web_fragments.fragment import Fragment
from xblock.core import XBlock
from xblock.fields import Boolean, Scope, String, UserScope
from xblock.utils.resources import ResourceLoader

from xblocks_contrib.common.xml_utils import LegacyXmlMixin, name_to_pathname

log = logging.getLogger(__name__)
resource_loader = ResourceLoader(__name__)

# The global (course-agnostic) anonymous user ID for the user.
ATTR_KEY_DEPRECATED_ANONYMOUS_USER_ID = "edx-platform.deprecated_anonymous_user_id"


class MLStripper(HTMLParser):
    "helper function for html_to_text below"

    def __init__(self):
        HTMLParser.__init__(self)
        self.reset()
        self.fed = []

    def handle_starttag(self, tag, attrs):
        if tag != "img":
            return
        for attr in attrs:
            if len(attr) >= 2 and attr[0] == "alt":
                self.fed.append(attr[1])

    def handle_data(self, data):
        """takes the data in separate chunks"""
        self.fed.append(data)

    def handle_entityref(self, name):
        """appends the reference to the body"""
        self.fed.append("&%s;" % name)

    def get_data(self):
        """joins together the seperate chunks into one cohesive string"""
        return "".join(self.fed)


def html_to_text(html):
    """Convert HTML to plain text."""
    htmlstripper = MLStripper()
    htmlstripper.feed(html)
    return htmlstripper.get_data()


def escape_html_characters(content):
    """
    Remove HTML characters that shouldn't be indexed using ElasticSearch indexer
    This method is complementary to html_to_text method found in xmodule/annotator_mixin.py

    Args:
        content (str): variable to escape html characters from

    Returns:
        content (str): content ready to be index by ElasticSearch

    """

    # Removing HTML comments
    return re.sub(
        r"<!--.*-->",
        "",
        # Removing HTML CDATA
        re.sub(
            r"<!\[CDATA\[.*\]\]>",
            "",
            # Removing HTML-encoded non-breaking space characters
            re.sub(r"(\s|&nbsp;|//)+", " ", html_to_text(content)),
        ),
    )


def check_html(html):
    """
    Check whether the passed in html string can be parsed by lxml.
    Return bool success.
    """
    parser = etree.HTMLParser()
    try:
        etree.fromstring(html, parser)
        return True
    except Exception:  # pylint: disable=broad-exception-caught
        pass
    return False


def stringify_children(node):
    """
    Return all contents of an xml tree, without the outside tags.
    e.g. if node is parse of

        "<html a="b" foo="bar">Hi <div>there <span>Bruce</span><b>!</b></div><html>"

    should return

        "Hi <div>there <span>Bruce</span><b>!</b></div>"

    fixed from
    http://stackoverflow.com/questions/4624062/get-all-text-inside-a-tag-in-lxml
    """
    # Useful things to know:

    # node.tostring() -- generates xml for the node, including start
    #                 and end tags.  We'll use this for the children.
    # node.text -- the text after the end of a start tag to the start
    #                 of the first child
    # node.tail -- the text after the end this tag to the start of the
    #                 next element.
    parts = [node.text]
    for c in node.getchildren():
        parts.append(etree.tostring(c, with_tail=True, encoding="unicode"))

    # filter removes possible Nones in texts and tails
    return "".join([part for part in parts if part])


@XBlock.needs("i18n")
# We 'want' the user service, but we don't strictly 'need' it.
# This makes our block more resilient. It won't crash in test environments
# where the user service might not be available.
@XBlock.wants("user")
class HtmlBlock(LegacyXmlMixin, XBlock):  # pylint: disable=abstract-method
    """
    The HTML XBlock.
    """

    # Indicates that this XBlock has been extracted from edx-platform.
    is_extracted = True
    display_name = String(
        display_name=_("Display Name"),
        help=_("The display name for this component."),
        scope=Scope.settings,
        # it'd be nice to have a useful default but it screws up other things; so,
        # use display_name_with_default for those
        default=_("Text"),
    )
    data = String(help=_("Html contents to display for this block"), default="", scope=Scope.content)
    upstream_data = String(
        help=_("Upstream html contents to store upstream data field"),
        default=None,
        hidden=True,
        enforce_type=True,
        scope=Scope.content,
    )
    source_code = String(
        help=_("Source code for LaTeX documents. This feature is not well-supported."), scope=Scope.settings
    )
    use_latex_compiler = Boolean(help=_("Enable LaTeX templates?"), default=False, scope=Scope.settings)
    editor = String(
        help=_(
            "Select Visual to enter content and have the editor automatically create the HTML. Select Raw to edit "
            "HTML directly. If you change this setting, you must save the component and then re-open it for editing."
        ),
        display_name=_("Editor"),
        default="visual",
        values=[{"display_name": _("Visual"), "value": "visual"}, {"display_name": _("Raw"), "value": "raw"}],
        scope=Scope.settings,
    )

    ENABLE_HTML_XBLOCK_STUDENT_VIEW_DATA = "ENABLE_HTML_XBLOCK_STUDENT_VIEW_DATA"

    uses_xmodule_styles_setup = True
    template_dir_name = "html"
    show_in_read_only_mode = True
    icon_class = "other"

    @property
    def category(self):
        return self.scope_ids.block_type

    @property
    def location(self):
        return self.scope_ids.usage_id

    @location.setter
    def location(self, value):
        assert isinstance(value, UsageKey)
        self.scope_ids = self.scope_ids._replace(
            def_id=value,  # Note: assigning a UsageKey as def_id is OK in old mongo / import system but wrong in split
            usage_id=value,
        )

    @property
    def url_name(self):
        return self.location.block_id

    @property
    def xblock_kvs(self):
        """
        Retrieves the internal KeyValueStore for this XModule.

        Should only be used by the persistence layer. Use with caution.
        """
        # if caller wants kvs, caller's assuming it's up to date; so, decache it
        self.save()
        return self._field_data._kvs  # pylint: disable=protected-access

    @XBlock.supports("multi_device")
    def student_view(self, _context):
        """Return a fragment that contains the html for the student view."""
        frag = Fragment(self.get_html())
        frag.add_css(resource_loader.load_unicode("static/css/html.css"))
        frag.add_javascript("""function HtmlBlock(runtime, element){}""")
        frag.initialize_js("HtmlBlock")
        return frag

    @XBlock.supports("multi_device")
    def public_view(self, context):
        """Returns a fragment that contains the html for the preview view."""
        return self.student_view(context)

    def student_view_data(self, context=None):  # pylint: disable=unused-argument
        """Return a JSON representation of the student_view of this XBlock."""
        if getattr(settings, "FEATURES", {}).get(self.ENABLE_HTML_XBLOCK_STUDENT_VIEW_DATA, False):
            return {"enabled": True, "html": self.get_html()}
        else:
            return {
                "enabled": False,
                "message": f'To enable, set FEATURES["{self.ENABLE_HTML_XBLOCK_STUDENT_VIEW_DATA}"]',
            }

    def get_html(self):
        """Returns html required for rendering the block."""
        if not self.data:
            return self.data

        data = self.data

        # This code is now more defensive.
        # We check if the user service and the user object exist before using them.
        # This prevents crashes if they are not available.
        user_service = self.runtime.service(self, "user")
        if user_service:
            current_user = user_service.get_current_user()
            if current_user:
                user_id = current_user.opt_attrs.get(ATTR_KEY_DEPRECATED_ANONYMOUS_USER_ID)
                if user_id:
                    data = data.replace("%%USER_ID%%", user_id)
                if current_user.emails:
                    email = current_user.emails[0]
                    data = data.replace("%%USER_EMAIL%%", email)

        # The course ID replacement is always safe to run.
        data = data.replace("%%COURSE_ID%%", str(self.scope_ids.usage_id.context_key))
        return data

    def studio_view(self, context=None):  # pylint: disable=unused-argument
        """Return a fragment that contains the html for the studio view."""
        # Only the ReactJS editor is supported for this block.
        # See https://github.com/openedx/frontend-app-authoring/tree/master/src/editors/containers/TextEditor
        raise NotImplementedError

    @classmethod
    def get_customizable_fields(cls) -> dict[str, str | None]:
        return {
            "display_name": "upstream_display_name",
            "data": "upstream_data",
        }

    # VS[compat] TODO (cpennington): Delete this method once all fall 2012 course
    # are being edited in the cms
    @classmethod
    def backcompat_paths(cls, filepath):
        """
        Get paths for html and xml files.
        """
        if filepath.endswith(".html.xml"):
            filepath = filepath[:-9] + ".html"  # backcompat--look for html instead of xml
        if filepath.endswith(".html.html"):
            filepath = filepath[:-5]  # some people like to include .html in filenames..
        candidates = []
        while os.sep in filepath:
            candidates.append(filepath)
            _, _, filepath = filepath.partition(os.sep)

        # also look for .html versions instead of .xml
        new_candidates = []
        for candidate in candidates:
            if candidate.endswith(".xml"):
                new_candidates.append(candidate[:-4] + ".html")
        return candidates + new_candidates

    @classmethod
    def filter_templates(cls, template, course):
        """
        Filter template that contains 'latex' from templates.

        Show them only if use_latex_compiler is set to True in
        course settings.
        """
        return "latex" not in template["template_id"] or course.use_latex_compiler

    def get_context(self):
        """
        an override to add in specific rendering context, in this case we need to
        add in a base path to our c4x content addressing scheme
        """
        # Add some specific HTML rendering context when editing HTML blocks where we pass
        # the root /c4x/ url for assets. This allows client-side substitutions to occur.
        return {
            "module": self,
            "editable_metadata_fields": self.editable_metadata_fields,  # pylint: disable=no-member
            "data": self.data,
            "base_asset_url": self.get_base_url_path_for_course_assets(self.location.course_key),
            "enable_latex_compiler": self.use_latex_compiler,
            "editor": self.editor,
        }

    def index_dictionary(self):
        xblock_body = super().index_dictionary()
        # Removing script and style
        html_content = re.sub(
            re.compile(
                r"""
                    <script>.*?</script> |
                    <style>.*?</style>
                """,
                re.DOTALL | re.VERBOSE,
            ),
            "",
            self.data,
        )
        html_content = escape_html_characters(html_content)
        html_body = {
            "html_content": html_content,
            "display_name": self.display_name,
        }
        if "content" in xblock_body:
            xblock_body["content"].update(html_body)
        else:
            xblock_body["content"] = html_body
        xblock_body["content_type"] = "Text"
        return xblock_body

    def bind_for_student(self, user_id, wrappers=None):
        """
        Set up this XBlock to act as an XModule instead of an XModuleDescriptor.

        Arguments:
            user_id: The user_id to set in scope_ids
            wrappers: These are a list functions that put a wrapper, such as
                      LmsFieldData or OverrideFieldData, around the field_data.
                      Note that the functions will be applied in the order in
                      which they're listed. So [f1, f2] -> f2(f1(field_data))
        """

        # Skip rebinding if we're already bound a user, and it's this user.
        if self.scope_ids.user_id is not None and user_id == self.scope_ids.user_id:
            if getattr(self.runtime, "position", None):
                # update the position of the tab
                self.position = self.runtime.position  # pylint: disable=attribute-defined-outside-init
            return

        # # If we are switching users mid-request, save the data from the old user.
        # self.save()

        # Update scope_ids to point to the new user.
        self.scope_ids = self.scope_ids._replace(user_id=user_id)

        # Clear out any cached instantiated children.
        self.clear_child_cache()

        # Clear out any cached field data scoped to the old user.
        for field in self.fields.values():
            if field.scope in (Scope.parent, Scope.children):
                continue

            if field.scope.user == UserScope.ONE:
                field._del_cached_value(self)  # pylint: disable=protected-access
                # not the most elegant way of doing this, but if we're removing
                # a field from the module's field_data_cache, we should also
                # remove it from its _dirty_fields
                if field in self._dirty_fields:
                    del self._dirty_fields[field]

        if wrappers:
            # Put user-specific wrappers around the field-data service for this block.
            # Note that these are different from modulestore.xblock_field_data_wrappers, which are not user-specific.
            wrapped_field_data = self.runtime.service(self, "field-data-unbound")
            for wrapper in wrappers:
                wrapped_field_data = wrapper(wrapped_field_data)
            self._bound_field_data = wrapped_field_data  # pylint: disable=attribute-defined-outside-init
            if getattr(self.runtime, "uses_deprecated_field_data", False):
                # This approach is deprecated but old mongo's CachingDescriptorSystem still requires it.
                # For Split mongo's CachingDescriptor system, don't set ._field_data this way.
                self._field_data = wrapped_field_data

    @staticmethod
    def serialize_asset_key_with_slash(asset_key):
        """
        Legacy code expects the serialized asset key to start w/ a slash; so, do that in one place
        :param asset_key:
        """
        url = str(asset_key)
        if not url.startswith("/"):
            url = "/" + url  # TODO - re-address this once LMS-11198 is tackled.
        return url

    @staticmethod
    def get_base_url_path_for_course_assets(course_key):  # pylint: disable=missing-function-docstring
        if (course_key is None) or isinstance(course_key, LibraryLocatorV2):
            return None

        assert isinstance(course_key, CourseKey)
        placeholder_id = uuid.uuid4().hex
        # create a dummy asset location with a fake but unique name. strip off the name, and return it
        url_path = HtmlBlock.serialize_asset_key_with_slash(
            course_key.make_asset_key("asset", placeholder_id).for_branch(None)
        )
        return url_path.replace(placeholder_id, "")

    def save(self):
        """Save block data."""
        if hasattr(self, "children"):
            self.children = [str(c) for c in self.children]
        super().save()

    @staticmethod
    def workbench_scenarios():
        """Create canned scenario for display in the workbench."""
        return [
            ("HtmlBlock", """<_html_extracted/>"""),
            (
                "Multiple HtmlBlock",
                """<vertical_demo><_html_extracted/><_html_extracted/><_html_extracted/></vertical_demo>""",
            ),
        ]

    # NOTE: html descriptors are special.  We do not want to parse and
    # export them ourselves, because that can break things (e.g. lxml
    # adds body tags when it exports, but they should just be html
    # snippets that will be included in the middle of pages.

    @classmethod
    def load_definition(cls, xml_object, system, def_id, id_generator):
        """Load a descriptor from the specified xml_object:

        If there is a filename attribute, load it as a string, and
        log a warning if it is not parseable by etree.HTMLParser.

        If there is not a filename attribute, the definition is the body
        of the xml_object, without the root tag (do not want <html> in the
        middle of a page)

        Args:
            xml_object: an lxml.etree._Element containing the definition to load
            system: the modulestore system or runtime which caches data
            location: the usage id for the block--used to compute the filename if none in the xml_object
            id_generator: used by other impls of this method to generate the usage_id
        """
        filename = xml_object.get("filename")
        if filename is None:
            definition_xml = copy.deepcopy(xml_object)
            cls.clean_metadata_from_xml(definition_xml)
            return {"data": stringify_children(definition_xml)}, []
        else:
            # html is special.  cls.filename_extension is 'xml', but
            # if 'filename' is in the definition, that means to load
            # from .html
            # 'filename' in html pointers is a relative path
            # (not same as 'html/blah.html' when the pointer is in a directory itself)
            pointer_path = "{category}/{url_path}".format(category="html", url_path=name_to_pathname(def_id.block_id))
            base = path(pointer_path).dirname()
            # log.debug("base = {0}, base.dirname={1}, filename={2}".format(base, base.dirname(), filename))
            filepath = f"{base}/{filename}.html"
            # log.debug("looking for html file for {0} at {1}".format(location, filepath))

            # VS[compat]
            # TODO (cpennington): If the file doesn't exist at the right path,
            # give the class a chance to fix it up. The file will be written out
            # again in the correct format.  This should go away once the CMS is
            # online and has imported all current (fall 2012) courses from xml
            if not system.resources_fs.exists(filepath):

                candidates = cls.backcompat_paths(filepath)
                # log.debug("candidates = {0}".format(candidates))
                for candidate in candidates:
                    if system.resources_fs.exists(candidate):
                        filepath = candidate
                        break
            try:
                with system.resources_fs.open(filepath, encoding="utf-8") as infile:
                    html = infile.read()
                    # Log a warning if we can't parse the file, but don't error
                    if len(html) > 0 and not check_html(html):
                        msg = f"Couldn't parse html in {filepath}"
                        log.warning(msg)
                        system.error_tracker("Warning: " + msg)

                    definition = {"data": html}

                    # TODO (ichuang): remove this after migration
                    # for Fall 2012 LMS migration: keep filename (and unmangled filename)
                    definition["filename"] = [filepath, filename]

                    return definition, []

            except ResourceNotFound as err:
                msg = "Unable to load file contents at path {}: {} ".format(filepath, err)
                # add more info and re-raise
                raise Exception(msg).with_traceback(sys.exc_info()[2])

    @classmethod
    def parse_xml_new_runtime(cls, node, runtime, keys):
        """
        Parse XML in the new learning-core-based runtime. Since it doesn't yet
        support loading separate .html files, the HTML data is assumed to be in
        a CDATA child or otherwise just inline in the OLX.
        """
        block = runtime.construct_xblock_from_class(cls, keys)
        block.data = stringify_children(node)
        # Attributes become fields.
        for name, value in node.items():
            cls._set_field_if_present(block, name, value, {})
        return block

    def definition_to_xml(self, resource_fs):
        """
        Returns an lxml Element representing the definition of this block.
        This version creates a self-contained definition file that includes
        all necessary metadata for a successful re-import.
        """

        # Write html to file, return an empty tag
        pathname = name_to_pathname(self.url_name)
        filepath = "{category}/{pathname}.html".format(category=self.category, pathname=pathname)

        resource_fs.makedirs(os.path.dirname(filepath), recreate=True)
        with resource_fs.open(filepath, "wb") as filestream:
            html_data = self.data.encode("utf-8")
            filestream.write(html_data)

        # write out the relative name
        relname = path(pathname).basename()

        elt = etree.Element("html")
        elt.set("filename", relname)
        return elt

    @property
    def non_editable_metadata_fields(self):
        """`use_latex_compiler` should not be editable in the Studio settings editor."""
        return super().non_editable_metadata_fields + [HtmlBlock.xml_attributes, HtmlBlock.use_latex_compiler]



"""Init for the xblocks_contrib package."""

from .annotatable import AnnotatableBlock
from .discussion import DiscussionXBlock
from .html import HtmlBlock
from .lti import LTIBlock
from .poll import PollBlock
from .problem import ProblemBlock
from .video import VideoBlock
from .word_cloud import WordCloudBlock

__version__ = "0.8.1"



"""
Minimal Django settings for running tests on xblocks_contrib.
"""

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
    }
}
